----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--WITH Ada.Text_IO;
WITH Ada.Exceptions;
WITH Ada.Streams;
WITH System;
WITH Non_Spark_Discrete_Random_Socket_Num_Pkg;
WITH Dns_Types;
USE TYPE dns_types.Unsigned_Short;
USE TYPE System.Bit_Order;
WITH Ada.Unchecked_Conversion;
with non_spark_command_line_pkg;
with flags;

--This code is all nonSPARK, and could raise tons of system-related
--exceptions if the host machine is having network problems.  They all
--relate to non-IRONSIDES issues (like inability to create a socket, port
--number in use, etc), but at some point it would be good to add
--some diagnostic output
PACKAGE BODY DNS_Network IS
   --# hide DNS_Network;
   TYPE Socket_Array_Type IS ARRAY (Socket_Array_Index_Type)
      OF Gnat.Sockets.Socket_Type;

   --GLOBAL STATE NEEDED TO MONITOR ALL SOCKETS

   --the socket for all downstream DNS traffic (formerly named
   --UDP_Socket)
   Dedicated_DNS_Downstream_Socket : Gnat.Sockets.Socket_Type;

   -- any socket that needs to be listened to must be placed here
   -- (this includes thye dedicated socket above)
   UDP_R_Socket_Set : Gnat.Sockets.Socket_Set_Type;

   -- The array of ephemeral sockets.  DNS messages from upstream servers
   -- are received on one of these sockets.  This array must be kept
   -- consistent with UDP_R_Socket_Set.
   UDP_R_Socket_Array : Socket_Array_Type;

   -- internal data structure required by gnat.sockets.check_selector()
   UDP_Socket_Selector : Gnat.Sockets.Selector_Type;

   --needed in Initialize_UDP_State below
   PROCEDURE Bind_Socket_To_Ephemeral_Local_Port (
         Socket : IN     Gnat.Sockets.Socket_Type) IS
      UDP_Address : Gnat.Sockets.Sock_Addr_Type;
      Port_Found  : Boolean;
      Port_Num    : dns_types.Unsigned_Short;
   BEGIN
      UDP_Address.Addr := Gnat.Sockets.ANY_INET_ADDR;
      Port_Found := False;
      --these loops can't be proven to terminate, do we care?
      WHILE NOT Port_Found LOOP
         --already have a 16-bit random value generator, use it
         Port_Num := Non_Spark_Discrete_Random_MessageID_Pkg.RandomID;

         --always true for now, included for readability/maintainability
         WHILE Port_Num < Ephemeral_Port_Min OR Port_Num > Ephemeral_Port_Max LOOP
            Port_Num := Non_Spark_Discrete_Random_MessageID_Pkg.RandomID;
         END LOOP;

         UDP_Address.Port := Gnat.Sockets.Port_Type(Port_Num);
         BEGIN
            -- this will raise an exception if the port is taken
            Gnat.Sockets.Bind_Socket(Socket,UDP_Address);
            Port_Found := True;
         EXCEPTION
            WHEN OTHERS =>
               NULL; --no big deal, just try another port
               --at some point, should add code that exits after some
               --reasonable number of attempts
               --Ada.Text_IO.Put_Line(Ada.Exceptions.Exception_Name(E) & ":  " & Ada.Exceptions.Exception_Message(E));
         END;  --exception block
      END LOOP;
   END Bind_Socket_To_Ephemeral_Local_Port;

   --build the global state needed to listen for DNS responses
   --from the upstream server
   --initializes UDP_Socket_Selector, UDP_R_Socket_Array, and adds
   --the standard UDP DNS socket to the set of sockets to listen to
   --(replaced Initialize_UDP)
   PROCEDURE Initialize_UDP_State IS
      UDP_Address : Gnat.Sockets.Sock_Addr_Type;
      Tmp_Socket: Gnat.Sockets.Socket_Type;
   BEGIN
      -- create the socket which will used for all UDP DNS queries
      -- from downstream clients
      Gnat.Sockets.Create_Socket(
         Socket => Dedicated_DNS_Downstream_Socket,
         Family => Gnat.Sockets.Family_Inet,
         Mode   => Gnat.Sockets.Socket_Datagram);

      -- Allow socket to be bound to address already in use
      Gnat.Sockets.Set_Socket_Option(Dedicated_DNS_Downstream_Socket,
         Gnat.Sockets.Socket_Level, (Gnat.Sockets.Reuse_Address, True));

      -- bind the socket to the first IP address on localhost, dns port 53
      UDP_Address.Addr := Gnat.Sockets.ANY_INET_ADDR;
      UDP_Address.Port := Gnat.Sockets.Port_Type(DNS_PORT);
      Gnat.Sockets.Bind_Socket(Dedicated_DNS_Downstream_Socket,UDP_Address);

      -- add it to the set of monitored sockets
      Gnat.Sockets.Set(UDP_R_Socket_Set, Dedicated_DNS_Downstream_Socket);

      --this is just an internal data structure, but you have to create one
      --if you're going to use Gnat.Sockets.Check_Selector()
      Gnat.Sockets.Create_Selector(UDP_Socket_Selector);

      --build the socket array
      --UDP_R_Socket_Array := (OTHERS => Gnat.Sockets.No_Socket);
      FOR Socket_Counter IN Socket_Array_Index_Type LOOP
         Gnat.Sockets.Create_Socket(
            Socket => Tmp_Socket,
            Family => Gnat.Sockets.Family_Inet,
            Mode   => Gnat.Sockets.Socket_Datagram);

         -- disallow ephemeral sockest to be bound to address already in use,
         -- since we need to make sure we find an ephemeral port not currently
         -- taken by some other application on our machine or another
         -- socket that we've opened previously
         Gnat.Sockets.Set_Socket_Option(Tmp_Socket,
            Gnat.Sockets.Socket_Level, (Gnat.Sockets.Reuse_Address, False));

         --bind it to a random local port to defend against cache poisoning
         Bind_Socket_To_Ephemeral_Local_Port(Tmp_Socket);

         --Add socket to UDP_R_Socket_Set and UDP_R_Socket_Array
         UDP_R_Socket_Array(Socket_Counter) := Tmp_Socket;
         Gnat.Sockets.Set(UDP_R_Socket_Set, Tmp_Socket);
      END LOOP;

   EXCEPTION
      WHEN E : OTHERS =>
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            Ada.Exceptions.Exception_Name(E) & ":  " &
            Ada.Exceptions.Exception_Message(E),0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
               "May need to change MAX_OPEN_SOCKETS in dns_network.ads",0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
   END Initialize_UDP_State;

   -------------------------
   -- Receive_DNS_Message --
   -------------------------
   PROCEDURE Receive_DNS_Message (
         Message       :    OUT DNS_Types.dns_message;
         Number_Bytes  :    OUT DNS_Types.message_length_range;
         Reply_Address :    OUT Network_Address_And_Port;
         Arriving_Port :    OUT Integer; --the port the message came in on
         Failure       :    OUT Boolean;
         Timeout       :    OUT Boolean) IS
      USE TYPE Dns_Types.message_length_range;
      USE TYPE Ada.Streams.Stream_Element_Offset;
      USE TYPE Gnat.Sockets.Selector_Status;
      USE TYPE Gnat.Sockets.Socket_Type;
      Last_Value                 : Ada.Streams.Stream_Element_Offset;
      UDP_Socket_Selector_Status : Gnat.Sockets.Selector_Status;
      UDP_W_Socket_Set           : Gnat.Sockets.Socket_Set_Type;                                          --ignored
      Tmp_Socket_Set             : Gnat.Sockets.Socket_Set_Type;
      Ready_Socket               : Gnat.Sockets.Socket_Type;
      ResponseReady              : boolean;
      Item                       : Ada.Streams.Stream_Element_Array (1 .. DNS_Types.dns_message'Size / 8);
      FOR Item'Address USE Message'Address;
      UPSTREAM_TIMEOUT_SEC : constant duration := 5.0;
   BEGIN
      Gnat.Sockets.Copy(
         Source => UDP_R_Socket_Set,
         Target => Tmp_Socket_Set);
      Gnat.Sockets.Check_Selector(
         Selector     => UDP_Socket_Selector,
         R_Socket_Set => Tmp_Socket_Set,
         W_Socket_Set => UDP_W_Socket_Set,
         --ignored
         Status  => UDP_Socket_Selector_Status,
         --Timeout => Gnat.Sockets.Forever);
         Timeout => UPSTREAM_TIMEOUT_SEC);
      Ready_Socket := Gnat.Sockets.No_Socket;
      Arriving_Port := 0;
      Timeout := False;
      ResponseReady := False;
      IF UDP_Socket_Selector_Status = Gnat.Sockets.Completed THEN
         --check for upstream responses first
         FOR Socket_Counter IN Socket_Array_Index_Type LOOP
            IF Gnat.Sockets.Is_Set(Tmp_Socket_Set,UDP_R_Socket_Array(Socket_Counter)) THEN
               Ready_Socket := UDP_R_Socket_Array(Socket_Counter);
               ResponseReady := True;
               --Ada.Text_IO.Put("Data ready on socket ");
               --Ada.Text_IO.Put(Integer'Image(Socket_Counter));
               --Ada.Text_IO.Put("port " & Integer'Image(Dns_Network.
                  --Port_Number_Of_Socket(dns_network.dns_socket(Ready_Socket))));
               --Ada.Text_IO.New_Line;
               EXIT;
            END IF;
         END LOOP;
         --if no responses, check for downstream queries
         IF not ResponseReady AND THEN
            Gnat.Sockets.Is_Set(Tmp_Socket_Set,
               Dedicated_DNS_Downstream_Socket) THEN
            -- message from downstream client
            --Ada.Text_IO.Put("Data ready on downstream socket ");
            --Ada.Text_IO.New_Line;
            Ready_Socket := Dedicated_DNS_Downstream_Socket;
         END IF; -- message from downstream client or upstream server

         --we know a message came in, so set the arriving port number
         Arriving_Port := Port_Number_Of_Socket(Dns_Network.Dns_Socket(Ready_Socket));

         Gnat.Sockets.Receive_Socket(
            Socket => Ready_Socket,
            Item   => Item,
            Last   => Last_Value,
            From   => Gnat.Sockets.Sock_Addr_Type (Reply_Address));
         Number_Bytes := DNS_Types.message_length_range(Last_Value);
         --put_line("Num bytes" & dns_types.message_length_range'image(number_bytes));
         IF System.Default_Bit_Order=System.Low_Order_First THEN
            DNS_Types.Byte_Swap(Message.Header);
         END IF;
         Failure := False;
      ELSE --socket selector status /= completed, nothing ready
         Failure := False;
         Timeout := True;
      END IF; --selector status

   EXCEPTION
      WHEN E: OTHERS =>
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,"Unexpected exception: ",0);
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
               Ada.Exceptions.Exception_Information(E),0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
         Failure := True;
   END Receive_DNS_Message;

   ---------------------------------
   -- Send_DNS_Message_Downstream --
   ---------------------------------
   PROCEDURE Send_DNS_Message_Downstream(
         Message : IN OUT DNS_Types.dns_message;
         Number_Bytes : IN     DNS_Types.message_length_range;
         To_Address   : IN     Network_Address_And_Port;
         Failure      :    OUT Boolean) IS
      USE TYPE Dns_Types.message_length_range;
      USE TYPE Ada.Streams.Stream_Element_Offset;
      Response_Stream : Ada.Streams.Stream_Element_Array (1 .. DNS_Types.dns_message'Size / 8);
      FOR Response_Stream'Address USE Message'Address;
      Result_Last : Ada.Streams.Stream_Element_Offset;
   BEGIN
      IF System.Default_Bit_Order=System.Low_Order_First THEN
         DNS_Types.Byte_Swap(Message.Header);
      END IF;

      Gnat.Sockets.Send_Socket(
         Socket => Dedicated_DNS_Downstream_Socket,
         Item   => Response_Stream (1 .. Ada.Streams.Stream_Element_Offset (Number_Bytes)),
         Last   => Result_Last,
         To     => Gnat.Sockets.Sock_Addr_Type (To_Address));
      --      put_line("Num bytes" & Ada.Streams.Stream_Element_Offset'image(result_last));
      IF Result_Last /= Ada.Streams.Stream_Element_Offset(Number_Bytes) THEN
         Failure := True;
      ELSE
         Failure := False;
      END IF;

   EXCEPTION
      WHEN E : OTHERS =>
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
             Ada.Exceptions.Exception_Name(E) & ":  " &
            Ada.Exceptions.Exception_Message(E),0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
         Failure := True;
   END Send_DNS_Message_Downstream;

   FUNCTION Get_Ephemeral_Socket RETURN Gnat.Sockets.Socket_Type IS
      Returned_Socket : Gnat.Sockets.Socket_Type;
   BEGIN
      BEGIN
         Returned_Socket :=
            UDP_R_Socket_Array(non_spark_discrete_random_socket_num_pkg.Random_Socket_Num);
      EXCEPTION
         WHEN E : OTHERS =>
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
             Ada.Exceptions.Exception_Name(E) & ":  " &
            Ada.Exceptions.Exception_Message(E),0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
      END;
      RETURN Returned_Socket;
   END Get_Ephemeral_Socket;

   FUNCTION Port_Number_Of_Socket (
         Socket : dns_socket)
     RETURN integer IS
   BEGIN
      RETURN integer(Gnat.Sockets.Get_Socket_Name(
         gnat.sockets.Socket_Type(Socket)).port);
   END Port_Number_Of_Socket;

   --not currently used
   FUNCTION Port_Number_Of_NAAP(
         Socket : network_address_and_port)
     RETURN Integer IS
   BEGIN
      RETURN integer(Socket.port);
   END Port_Number_Of_NAAP;


   -------------------------------
   -- Send_DNS_Message_Upstream --
   -------------------------------
   --like Send_DNS_Message_Downstream, but uses a dynamic ephemeral socket
   PROCEDURE Send_DNS_Message_Upstream (
         Message          : IN OUT DNS_Types.dns_message;
         Number_Bytes     : IN     DNS_Types.message_length_range;
         Upstream_Address : IN     Network_Address_And_Port;
         Outgoing_port    : OUT    integer;
         Failure          : OUT    Boolean) IS
      USE TYPE Dns_Types.message_length_range;
      USE TYPE Ada.Streams.Stream_Element_Offset;
      Tmp_Socket      : Gnat.Sockets.Socket_Type;
      Response_Stream : Ada.Streams.Stream_Element_Array (1 .. DNS_Types.dns_message'Size / 8);
      FOR Response_Stream'Address USE Message'Address;
      Result_Last : Ada.Streams.Stream_Element_Offset;
      loop_count : natural := 1;
   BEGIN
      IF System.Default_Bit_Order=System.Low_Order_First THEN
         DNS_Types.Byte_Swap(Message.Header);
      END IF;

      --Send to new socket with random ephemeral local port number
      Tmp_Socket := Get_Ephemeral_Socket;
      --Set the port number out parameter, needed by the caller
      Outgoing_Port :=  Port_Number_Of_Socket(Dns_Network.Dns_Socket(Tmp_Socket));

      if flags.remap then
         loop_count := 3;
      end if;
      -- Send the DNS message out on the wire using the new socket
      for i in integer range 1..loop_count loop
      Gnat.Sockets.Send_Socket(
         Socket => Tmp_Socket,
         Item   => Response_Stream (1 .. Ada.Streams.Stream_Element_Offset (Number_Bytes)),
         Last   => Result_Last,
            To     => Gnat.Sockets.Sock_Addr_Type (Upstream_Address));
      end loop;
      --      put_line("Num bytes" & Ada.Streams.Stream_Element_Offset'image(result_last));
      IF Result_Last /= Ada.Streams.Stream_Element_Offset(Number_Bytes) THEN
         Failure := True;
      ELSE
         Failure := False;
      END IF;

   EXCEPTION
      WHEN E : OTHERS =>
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
             Ada.Exceptions.Exception_Name(E) & ":  " &
            Ada.Exceptions.Exception_Message(E),0);
         Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
         Failure := True;
   END Send_DNS_Message_Upstream;

   PROCEDURE SetAddrAndPort (
         NAAP :    OUT Network_Address_And_Port;
         Addr : IN     String;
         Port : IN     Dns_Types.Unsigned_Short) IS
   BEGIN
      NAAP.Addr := Gnat.Sockets.Inet_Addr(Addr);
      NAAP.Port := Gnat.Sockets.Port_Type(Port);
   END SetAddrAndPort;

   PROCEDURE SetAddrAndPortCommandLine (
         NAAP :    OUT Network_Address_And_Port;
         Port : IN     Dns_Types.Unsigned_Short) IS
      TYPE String_Access IS ACCESS ALL String;
      FUNCTION to_string_access IS NEW Ada.Unchecked_Conversion(Non_Spark_Command_Line_Pkg.MyString,
         string_access);
   BEGIN
      NAAP.Addr := Gnat.Sockets.Inet_Addr(
         To_String_Access(Non_Spark_Command_Line_Pkg.Ip_To_Be_Used_String).all);
      NAAP.Port := Gnat.Sockets.Port_Type(Port);
   END SetAddrAndPortCommandLine;

   PROCEDURE Print_Server_Info (
         Server : IN     Network_Address_And_Port) IS
      Str : CONSTANT String := Gnat.Sockets.Image (Gnat.Sockets.Sock_Addr_Type (Server).Addr);
   BEGIN
      Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
         Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,Str,0);
      Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
         Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
   END Print_Server_Info;
END DNS_Network;
