----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
WITH DNS_Types;
with gnat.sockets;
WITH Protected_SPARK_IO_05;
WITH Non_Spark_Discrete_Random_MessageID_Pkg;
WITH ada.Real_Time;

--# inherit DNS_Types, Protected_SPARK_IO_05, System, ada.Real_Time;
package DNS_Network
--# own protected Network : Network_Type;
is
   --# type Network_Type is abstract;

   -- how long do we allow a receive/send to occur before we timeout
   Socket_Timeout_Milliseconds : CONSTANT := 2_000;

   DNS_PORT : CONSTANT dns_types.Unsigned_Short := 53;

   Ephemeral_Port_Min : CONSTANT dns_types.Unsigned_Short := 1024;
   Ephemeral_Port_Max : CONSTANT dns_types.Unsigned_Short := 65535;
   SUBTYPE Ephemeral_Port_Range IS dns_types.Unsigned_Short
      RANGE Ephemeral_Port_Min..Ephemeral_Port_Max;

   --the upper limit for this is OS dependent,
   --may need to be changed depending on platform
   MAX_OPEN_SOCKETS : CONSTANT natural := 768;
   SUBTYPE Socket_Array_Index_Type IS Natural RANGE 1..MAX_OPEN_SOCKETS;

   --how long a query is allowed to stay in the query buffer
   QUERY_TIME_LIMIT_SEC : CONSTANT Ada.Real_Time.Time_Span := Ada.Real_Time.Milliseconds(5000);

   --formerly limited private, but now we need to initialize it
   TYPE Network_Address_And_Port IS PRIVATE;
   --deferred constant, see private section below
   NO_NETWORK_ADDRESS_AND_PORT : Constant Network_Address_And_Port;
   type DNS_Socket is private;

   PROCEDURE Initialize_UDP_State;
   --# global in out Network;
   --# derives Network from *;

   -- Reads a single UDP DNS message from network
   -- Last is the number of bytes read (assuming no failure)
   -- post condition comes from DNS_Network_Receive

   PROCEDURE Receive_DNS_Message(
      Message : out DNS_Types.dns_message;
      Number_Bytes  : out DNS_Types.Message_Length_Range;
      Reply_Address : OUT Network_Address_And_Port;
      Arriving_Port : OUT Integer;
      Failure : OUT Boolean;
      Timeout : OUT Boolean);
   --# global in out Network;
   --# derives Failure, Timeout,
   --#         Network,
   --#         Number_Bytes,
   --#         Message,
   --#         Reply_Address, Arriving_Port from Network;

   -- send a single UDP DNS reply message to a downstream client,
   -- using the dedicated UDP DNS socket
   -- Last is number of bytes to send
   PROCEDURE Send_DNS_Message_Downstream(Message : in out DNS_Types.dns_message;
      Number_Bytes : in DNS_Types.Message_Length_Range;
      To_Address : in Network_Address_and_Port;
      Failure : out Boolean);
   --# global in out Network;
   --# derives Failure,
   --#         Network from Network,
   --#                      Number_Bytes,
   --#                      Message,
   --#                      To_Address &
   --#         Message  from Network;
   --# pre Integer(Number_Bytes) > DNS_Types.Header_Bits/8;
   --# post System.Default_Bit_Order=System.High_Order_First -> Message = Message~;

   -- send a single UDP DNS reply message to the given address of an upstream server
   -- using a randomly generated ephemeral socket
   -- Last is number of bytes to send
   PROCEDURE Send_DNS_Message_Upstream(Message : in out DNS_Types.dns_message;
      Number_Bytes : in DNS_Types.Message_Length_Range;
      Upstream_Address : IN Network_Address_And_Port;
      Outgoing_Port : out Integer;
      Failure : out Boolean);
   --# global in out Network;
   --# derives Failure, Network from Network,Number_Bytes, Message, Upstream_Address &
   --#         Message from Network &
   --#         Outgoing_Port from;
   --# pre Integer(Number_Bytes) > DNS_Types.Header_Bits/8;
   --# post System.Default_Bit_Order=System.High_Order_First -> Message = Message~;

   PROCEDURE Print_Server_Info(Server: IN Network_Address_And_Port);
   --# global in out Protected_SPARK_IO_05.SPARK_IO_PO;
   --# derives Protected_SPARK_IO_05.SPARK_IO_PO from *, Server;

   PROCEDURE SetAddrAndPort(NAAP : OUT Network_Address_And_Port;
      Addr : IN String; Port : IN DNS_Types.Unsigned_Short);
   --# derives NAAP from Addr, Port;

   PROCEDURE SetAddrAndPortCommandLine(NAAP : OUT Network_Address_And_Port;
      Port : IN DNS_Types.Unsigned_Short);
   --# derives NAAP from Port;

   FUNCTION Port_Number_Of_Socket (Socket : dns_socket)
      RETURN integer;

PRIVATE
   --# hide DNS_Network
   TYPE Network_Address_And_Port IS NEW Gnat.Sockets.Sock_Addr_Type;
   NO_NETWORK_ADDRESS_AND_PORT : CONSTANT Network_Address_And_Port :=
      Network_Address_And_Port'(Family => Gnat.Sockets.Family_Inet,
         Addr => Gnat.Sockets.No_Inet_Addr, Port => gnat.sockets.No_Port);
   type DNS_Socket is new Gnat.Sockets.Socket_Type;
end DNS_Network;
