----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--with ada.Text_IO;
package body Dns_Table_Pkg
IS
   FUNCTION Seconds_To_Timespan(Seconds: Integer) RETURN Ada.Real_Time.Time_Span IS
      --# hide Seconds_To_Timespan
   BEGIN
      return Ada.Real_Time.Seconds(Seconds);
   END Seconds_To_Timespan;

--DNS_Table is a protected object, these methods are atomic
   protected body Dns_Table_Type is
   --UTILITIES
   function To_Lower(DomainName : in Rr_Type.WireStringType) return Rr_Type.WireStringType is
      LowerCaseVersion : Rr_Type.WireStringType := Rr_Type.BlankWire;
      length : rr_type.wireStringTypeIndex;
   begin
      Length := Rr_Type.WireNameLength(domainName);
      for I in Integer range 1..Length loop
         --# assert true;
         LowerCaseVersion(I) := Ada.Characters.Handling.to_lower(domainName(I));
      end loop;
      return lowerCaseVersion;
   end To_Lower;

   function Same(X,Y : in RR_Type.WireStringType) return Boolean is
      Length : Rr_Type.WireStringTypeIndex;
      result : boolean;
   begin
      result := true;
      Length := Rr_Type.WireNameLength(x);
      for I in Integer range 1..Length loop
         --#assert true;
         if X(I)/=Y(I) then
            Result := False;
            exit;
         end if;
      end loop;
      return result;
   end Same;

   function hash(domainName : in rr_type.WireStringType)
      return Rr_Type.NumBucketsIndexType
      --# return val => (1 <= val and val <= rr_type.NumBuckets);
   is
      NumCharsInHashFunction : constant Natural := 4;
      val : natural := 0;
   begin
      for i in integer range 1..NumCharsInHashFunction loop
         --# assert val <= (i-1)*Character'Pos(Character'Last)
         --#   and (for all Q in rr_type.WireStringTypeIndex =>
         --#          (character'pos(domainname(Q))<=255 and
         --#           character'pos(domainname(Q))>=0));
         val := val + Character'Pos(domainName(i));
      end loop;
      return (val mod rr_type.NumBuckets) + 1;
   end hash;



   function ValidEntry(Stamp : in Ada.Real_Time.Time;
      TTLInSeconds : in Unsigned_Types.Unsigned32) return boolean
   --# global Ada.Real_Time.ClockTime;
   IS
      Now : Ada.Real_Time.Time;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
   begin
      --Have to do this to make SPARK happy, but not worth
      --printing a warning about.  TTL > 68 years treated as 68 years?
      --Who cares?
      if TTLInSeconds > Unsigned_Types.Unsigned32(Integer'Last) then
         TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer'Last);
      else
         TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTLInSeconds));
      end if;
      Now := Ada.Real_Time.Clock;
      return(Stamp + TimeSpanTTLInSeconds > Now);
   end ValidEntry;


--QUERY PROCEDURES
   procedure queryARecords(
   	domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.a_record_type.ARecordBucketType;
                           howMany : out rr_type.NumberOfRecordsType)
   --# global in ARecordTable, ARecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from ARecordKeys, ARecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from ARecordKeys, ARecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      Bucket : Rr_Type.NumBucketsIndexType;
      Lower_DomainName : Rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(DomainName);

      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_DomainName);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when ARecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(ARecordKeys(Bucket)(Ctr),Lower_DomainName) THEN
            valid := ValidEntry(ArecordTable(Bucket)(Ctr).TimeStamp,
               ArecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := ARecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if;
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   end QueryARecords;



      procedure queryCorruptRecords(
                                    domainName : in rr_type.WireStringType;
                                    inTable : out boolean)
      --# global in CorruptRecordKeys;
      --# derives intable from CorruptRecordKeys, domainName;
      is
         Bucket : Rr_Type.NumBucketsIndexType;
         valid : boolean;
         howMany : rr_type.NumberOfRecordsType;
      begin
         --queries must be case-insensitive, so hash on Lower case version
         Bucket := Hash(domainName);
         howMany := 0;
         for ctr in rr_type.ReturnedRecordsIndexType loop
            --# assert howMany >= 0 and howMany < ctr and bucket >= 1
            --#    and bucket <= rr_type.NumBuckets;
            exit when CorruptRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
            IF Same(CorruptRecordKeys(Bucket)(Ctr),domainName) THEN
                HowMany := HowMany+1;
            end if;
         end loop;

         if (howMany > 0) THEN
            inTable := true;
         else
            inTable := false;
         end if;
      end QueryCorruptRecords;



   procedure queryAAAARecords(
   	domainName : in rr_type.WireStringType;
      returnedRecords : out rr_type.aaaa_record_type.AAAARecordBucketType;
      howMany : out rr_type.NumberOfRecordsType)
   --# global in AAAARecordTable, AAAARecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from AAAARecordKeys, AAAARecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from AAAARecordKeys, AAAARecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_DomainName);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when AAAARecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(AAAARecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(AAAArecordTable(Bucket)(Ctr).TimeStamp,
               AAAArecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := AAAARecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   end queryAAAARecords;


   procedure queryCNAMERecords(domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.cname_record_type.CNAMERecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in CNAMERecordTable, CNAMERecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from CNAMERecordKeys, CNAMERecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from CNAMERecordKeys, CNAMERecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when CNAMERecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(CNAMERecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(CNAMErecordTable(Bucket)(Ctr).TimeStamp,
               CNAMErecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := CNAMERecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QueryCNAMERecords;

   procedure queryDNSKEYRecords(domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.DNSKEY_record_type.DNSKEYRecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in DNSKEYRecordTable, DNSKEYRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from DNSKEYRecordKeys, DNSKEYRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from DNSKEYRecordKeys, DNSKEYRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when DNSKEYRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(DNSKEYRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(DNSKEYrecordTable(Bucket)(Ctr).TimeStamp,
               DNSKEYrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := DNSKEYRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;--# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QueryDNSKEYRecords;

   procedure queryMXRecords(
      domainName : in rr_type.WireStringType;
      returnedRecords : out rr_type.mx_record_type.MXRecordBucketType;
      howMany : out rr_type.NumberOfRecordsType)
   --# global in MXRecordTable, MXRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from MXRecordKeys, MXRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from MXRecordKeys, MXRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);

      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when MXRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(MXRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(MXrecordTable(Bucket)(Ctr).TimeStamp,
               MXrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := MXRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;--# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END queryMXRecords;

   procedure queryNSRecords(
   	domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.ns_record_type.NSRecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in NSRecordTable, NSRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from NSRecordKeys, NSRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from NSRecordKeys, NSRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);

      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when NSRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(NSRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(NSrecordTable(Bucket)(Ctr).TimeStamp,
               NSrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := NSRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   end QueryNSRecords;

   --SPARC can't prove it, but because of the precondition in insertNSECRecords,
   --all the records that come back from this query have numberOfBlocks > 0
   procedure queryNSECRecords(
   	domainName : in rr_type.WireStringType;
    returnedRecords : out rr_type.nsec_record_type.NSECRecordBucketType;
    howMany : out rr_type.NumberOfRecordsType)
   --# global in NSECRecordTable, NSECRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from NSECRecordKeys, NSECRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from NSECRecordKeys, NSECRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when NSECRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(NSECRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(NSECrecordTable(Bucket)(Ctr).TimeStamp,
               NSECrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := NSECRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   end queryNSECRecords;

procedure queryNXDOMAINRecords(
   	domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.NXDOMAIN_record_type.NXDOMAINRecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in NXDOMAINRecordTable, NXDOMAINRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from NXDOMAINRecordKeys, NXDOMAINRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from NXDOMAINRecordKeys, NXDOMAINRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
     --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);

      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when NXDOMAINRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(NXDOMAINRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(NXDOMAINrecordTable(Bucket)(Ctr).TimeStamp,
               NXDOMAINrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := NXDOMAINRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QueryNXDOMAINRecords;

   procedure queryPTRRecords(
   	domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.PTR_record_type.PTRRecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in PTRRecordTable, PTRRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from PTRRecordKeys, PTRRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#         returnedRecords from PTRRecordKeys, PTRRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
     --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);

      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when PTRRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(PTRRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(PTRrecordTable(Bucket)(Ctr).TimeStamp,
               PTRrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := PTRRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QueryPTRRecords;

   procedure queryRRSIGRecords(domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.RRSIG_record_type.RRSIGRecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in RRSIGRecordTable, RRSIGRecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from RRSIGRecordKeys, RRSIGRecordTable, domainName, Ada.Real_Time.ClockTime &
   --#    returnedRecords from RRSIGRecordKeys, RRSIGRecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);

      --queries must be case-insensitive, so hash on lower case version
      Bucket := Hash(Lower_Domainname);
      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when RRSIGRecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(RRSIGRecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(RRSIGrecordTable(Bucket)(Ctr).TimeStamp,
               RRSIGrecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := RRSIGRecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QueryRRSIGRecords;

   procedure querySOARecords(
   	domainName : in rr_type.WireStringType;
        returnedRecords : out rr_type.SOA_record_type.SOARecordBucketType;
        howMany : out rr_type.NumberOfRecordsType)
   --# global in SOARecordTable, SOARecordKeys, Ada.Real_Time.ClockTime;
   --# derives howMany from SOARecordKeys, SOARecordTable, domainName, Ada.Real_Time.ClockTime &
   --#    returnedRecords from SOARecordKeys, SOARecordTable, domainName, Ada.Real_Time.ClockTime;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Domainname : rr_Type.WireStringType;
      valid : boolean;
   begin
      Lower_Domainname := To_Lower(Domainname);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Domainname);

      howMany := 0;
      for ctr in rr_type.ReturnedRecordsIndexType loop
         --# assert howMany >= 0 and howMany < ctr and bucket >= 1
         --#    and bucket <= rr_type.NumBuckets;
         exit when SOARecordKeys(Bucket)(Ctr)(1) = ASCII.NUL;
         IF Same(SOARecordKeys(Bucket)(Ctr),Lower_DomainName) then
	        valid := ValidEntry(SOArecordTable(Bucket)(Ctr).TimeStamp,
               SOArecordTable(Bucket)(Ctr).TTLInSeconds);
            if valid then
               HowMany := HowMany+1;
               --# accept Flow, 23, returnedRecords, "assigning to uninitialized array is OK";
               returnedRecords(howMany) := SOARecordTable(bucket)(ctr);
               --# end accept;
            end if; --matching entry found with valid time stamp
         end if; --matching entry found
      end loop;
      --# accept Flow, 602, returnedRecords, returnedRecords, "fills as much of array as needed";
   END QuerySOARecords;

--INSERT PROCEDURES
   procedure InsertARecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.a_record_type.ARecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out ARecordTable, ARecordKeys;
   --# derives ARecordTable from *, Ada.Real_Time.ClockTime, ARecordKeys, theRecord, Key &
   --#         ARecordKeys from *, theRecord, ARecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);
      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if ARecordKeys(bucket)(i) = rr_type.BlankOwner then
            ARecordKeys(bucket)(i) := Lower_Key;
            ARecordTable(Bucket)(I) := TheRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := ARecordTable(Bucket)(I).TimeStamp;
            TTL := ARecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         ARecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         ARecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
      END InsertARecord;

   procedure InsertCorruptRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.corrupt_record_type.CorruptRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out CorruptRecordTable, CorruptRecordKeys;
   --# derives CorruptRecordTable from *, Ada.Real_Time.ClockTime, CorruptRecordKeys, theRecord, Key &
   --#         CorruptRecordKeys from *, theRecord, CorruptRecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

     --Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output, "Inserting New Domain", 0);
     --Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output, Lower_Key, 0);


      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if CorruptRecordKeys(bucket)(i) = rr_type.BlankOwner then
            CorruptRecordKeys(bucket)(i) := Lower_Key;
            CorruptRecordTable(Bucket)(I) := TheRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := CorruptRecordTable(Bucket)(I).TimeStamp;
            TTL := CorruptRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         CorruptRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         CorruptRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertCorruptRecord;

   procedure insertAAAARecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.aaaa_record_type.AAAARecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out AAAARecordTable, AAAARecordKeys;
   --# derives AAAARecordTable from *, Ada.Real_Time.ClockTime, AAAARecordKeys, theRecord, Key &
   --#         AAAARecordKeys from *, theRecord, AAAARecordTable,Ada.Real_Time.ClockTime, Key;
   is
      Bucket : Rr_Type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;

   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if AAAARecordKeys(bucket)(i) = rr_type.BlankOwner then
            AAAARecordKeys(bucket)(i) := Lower_Key;
            AAAARecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := AAAARecordTable(Bucket)(I).TimeStamp;
            TTL := AAAARecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         AAAARecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         AAAARecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertAAAARecord;

   procedure InsertCNAMERecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.cname_record_type.CNAMERecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out CNAMERecordTable, CNAMERecordKeys;
   --# derives CNAMERecordTable from *, Ada.Real_Time.ClockTime, CNAMERecordKeys, theRecord, Key &
   --#         CNAMERecordKeys from *, theRecord, CNAMERecordTable,Ada.Real_Time.ClockTime, Key;
      is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;

   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;

         --CNAME records are special.  It's possible that records could
         --arrive corresponding to matching entries in the cache if
         --a query is being resent due to a CNAME chain with expired
         --records at the end.  Overwrite with the new one, so that the
         --cache never has more than one CNAME record for the same domain name.
         IF Same(CNAMERecordKeys(Bucket)(I),Lower_Key) then
            CNAMERecordKeys(Bucket)(I) := Rr_Type.BlankOwner;
         end if;

         if CNAMERecordKeys(bucket)(i) = rr_type.BlankOwner then
            CNAMERecordKeys(bucket)(i) := Lower_Key;
            CNAMERecordTable(bucket)(i) := theRecord;
            Success := True;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := CNAMERecordTable(Bucket)(I).TimeStamp;
            TTL := CNAMERecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         CNAMERecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         CNAMERecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;

   END InsertCNAMERecord;

   procedure InsertDNSKEYRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.DNSKEY_record_type.DNSKEYRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#   in out DNSKEYRecordTable, DNSKEYRecordKeys;
   --# derives DNSKEYRecordTable from *, Ada.Real_Time.ClockTime, DNSKEYRecordKeys, theRecord, Key &
   --#         DNSKEYRecordKeys from *, theRecord, DNSKEYRecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if DNSKEYRecordKeys(bucket)(i) = rr_type.BlankOwner then
            DNSKEYRecordKeys(bucket)(i) := Lower_Key;
            DNSKEYRecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;
      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := DNSKEYRecordTable(Bucket)(I).TimeStamp;
            TTL := DNSKEYRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         DNSKEYRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         DNSKEYRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertDNSKEYRecord;

   procedure InsertMXRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.MX_record_type.MXRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out MXRecordTable, MXRecordKeys;
   --# derives MXRecordTable from *, Ada.Real_Time.ClockTime, MXRecordKeys, theRecord, Key &
   --#         MXRecordKeys from *, theRecord, MXRecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if MXRecordKeys(bucket)(i) = rr_type.BlankOwner then
            MXRecordKeys(bucket)(i) := Lower_Key;
            MXRecordTable(Bucket)(I) := TheRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := MXRecordTable(Bucket)(I).TimeStamp;
            TTL := MXRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         MXRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         MXRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;

   END InsertMXRecord;

   procedure insertNSRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.ns_record_type.NSRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out NSRecordTable, NSRecordKeys;
   --# derives NSRecordTable from *, Ada.Real_Time.ClockTime, NSRecordKeys, theRecord, Key &
   --#         NSRecordKeys from *, theRecord, NSRecordTable,Ada.Real_Time.ClockTime, Key;
  is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if NSRecordKeys(bucket)(i) = rr_type.BlankOwner then
            NSRecordKeys(bucket)(i) := Lower_Key;
            NSRecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := NSRecordTable(Bucket)(I).TimeStamp;
            TTL := NSRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         NSRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         NSRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertNSRecord;

   procedure insertNSECRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.nsec_record_type.NSECRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out NSECRecordTable, NSECRecordKeys;
   --# derives NSECRecordTable from *, Ada.Real_Time.ClockTime, NSECRecordKeys, theRecord, Key &
   --#         NSECRecordKeys from *, theRecord, NSECRecordTable, Ada.Real_Time.ClockTime,  Key;
  is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if NSECRecordKeys(bucket)(i) = rr_type.BlankOwner then
            NSECRecordKeys(bucket)(i) := Lower_Key;
            NSECRecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;
      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := NSECRecordTable(Bucket)(I).TimeStamp;
            TTL := NSECRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         NSECRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         NSECRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertNSECRecord;

   procedure InsertNXDomainRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.nxdomain_record_type.nxdomainRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out NXDOMAINRecordTable, NXDOMAINRecordKeys;
   --# derives NXDOMAINRecordTable from *, Ada.Real_Time.ClockTime, NXDOMAINRecordKeys, theRecord, Key &
   --#         NXDOMAINRecordKeys from *, theRecord, NXDOMAINRecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if NXDOMAINRecordKeys(bucket)(i) = rr_type.BlankOwner then
            NXDOMAINRecordKeys(bucket)(i) := Lower_Key;
            NXDOMAINRecordTable(Bucket)(I) := TheRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := NXDOMAINRecordTable(Bucket)(I).TimeStamp;
            TTL := NXDOMAINRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         NXDOMAINRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         NXDOMAINRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertNXDomainRecord;

   procedure insertPTRRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.PTR_record_type.PTRRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#    in out PTRRecordTable, PTRRecordKeys;
   --# derives PTRRecordTable from *, Ada.Real_Time.ClockTime, PTRRecordKeys, theRecord, Key &
   --#         PTRRecordKeys from *, theRecord, PTRRecordTable,Ada.Real_Time.ClockTime, Key;
  is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;
   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if PTRRecordKeys(bucket)(i) = rr_type.BlankOwner then
            PTRRecordKeys(bucket)(i) := Lower_Key;
            PTRRecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := PTRRecordTable(Bucket)(I).TimeStamp;
            TTL := PTRRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         PTRRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         PTRRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;

   END InsertPTRRecord;

   procedure InsertRRSIGRecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.rrsig_record_type.rrsigRecordType)
   --# global in Ada.Real_Time.ClockTime;
   --#   in out RRSIGRecordTable, RRSIGRecordKeys;
   --# derives RRSIGRecordTable from *, Ada.Real_Time.ClockTime, RRSIGRecordKeys, theRecord, Key &
   --#         RRSIGRecordKeys from *, theRecord, RRSIGRecordTable,Ada.Real_Time.ClockTime, Key;
   is
      bucket : rr_type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;

   begin
      Lower_Key := To_Lower(Key);
      --queries must be case-insensitive, so hash on Lower case version
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if rrsigRecordKeys(bucket)(i) = rr_type.BlankOwner then
            rrsigRecordKeys(bucket)(i) := Lower_Key;
            rrsigRecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;
      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := RRSIGRecordTable(Bucket)(I).TimeStamp;
            TTL := RRSIGRecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         RRSIGRecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         RRSIGRecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertRRSIGRecord;

   procedure insertSOARecord(Key : in Rr_Type.WireStringType;
                          theRecord : in rr_type.SOA_record_type.SOARecordType)
   --# global in out SOARecordTable, SOARecordKeys;
   --#        in Ada.Real_Time.Clocktime;
   --# derives SOARecordTable from *, SOARecordKeys, theRecord, Key, Ada.Real_Time.Clocktime  &
   --#         SOARecordKeys from *, SOARecordTable, theRecord, Ada.Real_Time.Clocktime, Key;
   is
      Bucket : Rr_Type.NumBucketsIndexType;
      Lower_Key : RR_Type.WireStringType;
      success : boolean;
      Stamp : Ada.Real_Time.Time;
      TTL:  Unsigned_types.unsigned32;
      TimeSpanTTLInSeconds : Ada.Real_Time.Time_Span;
      valid : boolean;
      ReplacementSlot : Natural;
      OldestSlot : Natural;
      Oldest : Ada.Real_Time.Time;

   begin
      Lower_Key := To_Lower(Key);
      Bucket := Hash(Lower_Key);

      success := false;
      for i in rr_type.ReturnedRecordsIndexType loop
         --# assert true;
         if SOARecordKeys(bucket)(i) = rr_type.BlankOwner then
            SOARecordKeys(bucket)(i) := Lower_Key;
            SOARecordTable(bucket)(i) := theRecord;
            success := true;
         end if;
         exit when success;
      END LOOP;

      IF NOT Success THEN
         --bucket full, need to replace an entry
         --if an entry has expired, replace that one, otherwise
         --replace the entry with the lowest Stamp+TTL
         Oldest := Ada.Real_Time.Time_Last;
         ReplacementSlot := 0;
         OldestSlot := 1;
         FOR I IN Rr_Type.ReturnedRecordsIndexType LOOP
            --# assert OldestSlot >= 1 and OldestSlot <= I and
            --#    ReplacementSlot <= I and I <= Rr_Type.ReturnedRecordsindexType'last;
            Stamp := SOARecordTable(Bucket)(I).TimeStamp;
            TTL := SOARecordTable(Bucket)(I).TTLInSeconds;
            valid := ValidEntry(Stamp,TTL);
            IF not valid THEN
               ReplacementSlot := I;
            ELSE
               if TTL <= Unsigned_Types.Unsigned32(Integer'Last) then
                  TimeSpanTTLInSeconds := Seconds_To_Timespan(Integer(TTL));
                  if Stamp + TimeSpanTTLInSeconds < Oldest THEN
                     OldestSlot := I;
                  end if;
               end if;
            end if;
            exit when ReplacementSlot > 0;
         END LOOP;
         IF ReplacementSlot = 0 THEN
            ReplacementSlot := OldestSlot;
         end if;
         SOARecordKeys(bucket)(ReplacementSlot) := Lower_Key;
         SOARecordTable(Bucket)(ReplacementSlot) := TheRecord;
      END IF;
   END InsertSOARecord;

END Dns_Table_Type;

end dns_table_pkg;
