----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--with Ada.Text_Io;
WITH DNS_Types;
WITH Unsigned_Types;
WITH System;
With DNS_Network;
use type DNS_Types.message_bytes_range;
use type DNS_Types.message_length_range;
use type DNS_Types.Byte;
use type DNS_Types.Query_Type;
USE TYPE DNS_Types.Query_Class;
use type Unsigned_Types.Unsigned8;
use type Unsigned_Types.Unsigned16;
use type Unsigned_Types.Unsigned32;
use type DNS_Types.Unsigned_Short;
USE TYPE DNS_Types.QNAME_PTR_RANGE;
use type DNS_Types.Response_Code;
USE TYPE System.Bit_Order;
USE TYPE DNS_Network.Network_Address_And_Port;
with flags;
PACKAGE BODY Process_Dns_Request IS


   function To_Lower(DomainName : in Rr_Type.WireStringType) return Rr_Type.WireStringType is
      LowerCaseVersion : Rr_Type.WireStringType := Rr_Type.BlankWire;
      length : rr_type.wireStringTypeIndex;
   begin
      Length := Rr_Type.WireNameLength(domainName);
      for I in Integer range 1..Length loop
         --# assert true;
         LowerCaseVersion(I) := Ada.Characters.Handling.to_lower(domainName(I));
      end loop;
      return lowerCaseVersion;
   END To_Lower;

   procedure Set_Unsigned_32(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         Value : in Unsigned_Types.Unsigned32) is
   begin
      Bytes(Start_Byte) := DNS_Types.Byte(Value/2**24);
      Bytes(Start_Byte+1) := DNS_Types.Byte((Value/2**16) mod 256);
      Bytes(Start_Byte+2) := DNS_Types.Byte((Value/2**8) mod 256);
      Bytes(Start_Byte+3) := DNS_Types.Byte(Value mod 256);
   END Set_Unsigned_32;

   procedure Set_Unsigned_16(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         Value : in Unsigned_Types.Unsigned16) is
   begin
      Bytes(Start_Byte) := DNS_Types.Byte((Value/2**8) mod 256);
      Bytes(Start_Byte+1) := DNS_Types.Byte(Value mod 256);
   end Set_Unsigned_16;

   --base64 conversion
   --implements Base64 index table at http://en.wikipedia.org/wiki/Base64
   --Arg is the character, ans is the value
   FUNCTION Base64CharToVal(Arg: IN Character) RETURN DNS_Types.Byte IS
      ans : DNS_Types.Byte;
   BEGIN
      CASE Arg IS
         WHEN 'A'..'Z' => ans := Character'Pos(Arg) - Character'Pos('A');
         WHEN 'a'..'z' => ans := (Character'Pos(Arg) - Character'Pos('a'))+26;
         WHEN '0'..'9' => ans := (Character'Pos(Arg) - Character'Pos('0'))+52;
         WHEN '+' => ans := 62;
         WHEN '/' => ans := 63;
         WHEN OTHERS => ans := 0;
      end case;
      return ans;
   END Base64CharToVal;

   procedure Set_TTL_Data_A_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         A_Record : in Rr_Type.A_Record_Type.ARecordType) is
   begin
         -- TTL (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte,A_Record.ttlInSeconds);
         -- DATA (2 bytes)
         Set_Unsigned_16(Bytes,Start_Byte+4,4);
         -- IP address
         Set_Unsigned_32(Bytes,Start_Byte+6,A_Record.ipv4);
   END Set_TTL_Data_A_Response;

   procedure Set_TTL_Data_AAAA_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         AAAA_Record : in Rr_Type.AAAA_Record_Type.AAAARecordType) is
   begin
      -- TTL (4 bytes)
      Set_Unsigned_32(Bytes,Start_Byte,AAAA_Record.ttlInSeconds);
      -- DATA (16 bytes)
      Set_Unsigned_16(Bytes,Start_Byte+4,16);
      for i in rr_type.aaaa_record_type.IPV6AddrTypeIndex loop
         --# assert true;
         -- IP address
         Set_Unsigned_16(Bytes,Start_Byte+
            DNS_Types.message_bytes_range(6+2*(I-Rr_Type.Aaaa_Record_Type.IPV6AddrTypeIndex'First)),
            AAAA_Record.Ipv6(i));
      end loop;
   end Set_TTL_Data_AAAA_Response;

   --see http://rfc-ref.org/RFC-TEXTS/4034/chapter2.html for DNSKEY record format
   procedure Set_TTL_Data_DNSKEY_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         DNSKEY_Record : IN Rr_Type.DNSKEY_Record_Type.DNSKEYRecordType;
         dstbytes : out rr_type.dnskey_record_type.keyLengthValueType) IS
         srcByte : Integer;
         KeyLength : Rr_Type.Dnskey_Record_Type.KeyLengthValueType;
         SrcNum0, SrcNum1, SrcNum2, SrcNum3: Dns_Types.Byte;
         dstNum0, dstNum1, dstNum2 : dns_types.byte;
   begin
         -- TTL (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte,DNSKEY_Record.ttlInSeconds);
         -- RDLENGTH (2 bytes, value is 4 plus the length of the key, placeholder at this point)
         Set_Unsigned_16(Bytes,Start_Byte+4,4+Unsigned_Types.Unsigned16(DNSKEY_Record.keyLength));
         -- RDATA (4 bytes)
         --   flags (2 bytes)
         Set_Unsigned_16(Bytes,Start_Byte+6,DNSKEY_Record.Flags);
         --   protocol (1 byte)
         Bytes(Start_Byte+8) := DNS_types.Byte(DNSKEY_Record.Protocol);
         --   algorithm (1 byte)
         Bytes(Start_Byte+9) := DNS_types.Byte(DNSKEY_Record.Algorithm);
         --   key field
         srcByte := 1;
         dstBytes := 1;
         keyLength := DNSKEY_Record.KeyLength;
         while srcByte + 3 <= keyLength loop  --since it's base64 encoded, src key will always contain an exact multiple of 4 bytes
            --# assert srcByte + 3 <= KeyLength and
            --# KeyLength <= rr_type.DNSKEY_Record_Type.MaxDNSKeyLength and KeyLength = DNSKEY_Record.KeyLength and
            --# dstBytes >= 1 and dstBytes <= srcByte and
            --# Start_Byte <= DNS_Types.message_bytes_range'Last-10-DNS_Types.message_bytes_range(DNSKEY_Record.keyLength);
            srcNum0 := Base64CharToVal(DNSKEY_Record.Key(srcByte));     --00aaaaaa
            srcNum1 := Base64CharToVal(DNSKEY_Record.Key(srcByte+1));   --00bbbbbb
            srcNum2 := Base64CharToVal(DNSKEY_Record.Key(srcByte+2));   --00cccccc
            SrcNum3 := Base64CharToVal(DNSKEY_Record.Key(SrcByte+3));   --00dddddd

            DstNum0 := SrcNum0*4 + (Srcnum1/16);   --aaaaaa00 + 000000bb = aaaaaabb
            DstNum1 := Srcnum1*16 + (SrcNum2/4);   --bbbb0000 + 0000cccc = bbbbcccc
            DstNum2 := SrcNum2*64 + SrcNum3;       --cc000000 + 00dddddd = ccdddddd

            Bytes((Start_Byte+9)+DNS_Types.message_bytes_range(DstBytes)) := DstNum0;
            Bytes((Start_Byte+9)+DNS_Types.message_bytes_range(DstBytes+1)) := DstNum1;
            Bytes((Start_Byte+9)+DNS_Types.message_bytes_range(DstBytes+2)) := DstNum2;

            srcByte := srcByte + 4;
            DstBytes := DstBytes + 3;
         END LOOP;
         -- RDLENGTH (2 bytes, value is 4 plus the length of the key, which we now know)
         dstBytes := dstbytes-1; --this is also an out parameter of this procedure
         Set_Unsigned_16(Bytes,Start_Byte+4,4+Unsigned_Types.Unsigned16(dstBytes));
   END Set_TTL_Data_DNSKEY_Response;

procedure Set_TTL_Data_MX_Response(
         Bytes               : in out DNS_Types.Bytes_Array_Type;
         Start_Byte          : in DNS_Types.message_bytes_range;
         MX_Record           : in Rr_Type.MX_record_type.MXRecordType;
         Current_Name_Length : in RR_Type.WireStringTypeIndex) is
   begin
         -- TTL
         Set_Unsigned_32(Bytes,Start_Byte,MX_Record.ttlInSeconds);
         -- DATA # bytes is equal to length of WireString + 2, -1 for removed null byte
         Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16((Current_Name_Length+2)-1));
         -- MAIL exchanger preference
         Set_Unsigned_16(Bytes,Start_Byte+6,MX_Record.pref);
         -- copy NS record
         for i in RR_Type.WireStringTypeIndex range 1..Current_Name_Length loop
            --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-8)-DNS_Types.message_bytes_range(Current_Name_Length) and
            --# i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last;
            Bytes((Start_Byte+7)+DNS_Types.message_bytes_range(I)) := DNS_Types.Byte(
               Character'Pos(MX_Record.mailExchanger(i)));
         end loop;
   end Set_TTL_Data_MX_Response;

 procedure Set_TTL_Data_NS_Response(
         Bytes               : in out DNS_Types.Bytes_Array_Type;
         Start_Byte          : in DNS_Types.message_bytes_range;
         NS_Record           : in Rr_Type.ns_record_type.NSRecordType;
         Current_Name_Length : in RR_Type.WireStringTypeIndex) is
   begin
         -- TTL
         Set_Unsigned_32(Bytes,Start_Byte,NS_Record.ttlInSeconds);
         -- DATA # bytes is equal to length of WireString
         Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16(Current_name_Length));
         -- copy NS record
         for i in RR_Type.WireStringTypeIndex range 1..Current_Name_Length loop
            --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-6)-DNS_Types.message_bytes_range(Current_Name_Length) and
            --# i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last;
            Bytes((Start_Byte+5)+DNS_Types.message_bytes_range(I)) := DNS_Types.Byte(
               Character'Pos(NS_Record.nameServer(i)));
         end loop;
   END Set_TTL_Data_NS_Response;

--see http://tools.ietf.org/rfc/rfc3845.txt for NSEC wire format.  If you're a masochist.
   procedure Set_TTL_Data_NSEC_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         NSEC_Record : IN Rr_Type.NSEC_Record_Type.NSECRecordType;
         Dstbytes : OUT DNS_Types.message_length_range) IS --bytes used in RRDATA section
         Current_Name_Length : RR_Type.WireStringTypeIndex;
         blockOffset : DNS_Types.message_bytes_range;
   BEGIN
      -- TTL (4 bytes)
      Set_Unsigned_32(Bytes,Start_Byte,NSEC_Record.TtlInSeconds);
      -- RDLENGTH (2 bytes, value will be the length of the domain name plus the length of the bitmapped RRList)
      -- placeholder for now
      Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16(0));

      -- NEXT DOMAIN NAME (Current_Name_Length bytes)
      Current_Name_Length := RR_Type.WireNameLength(NSEC_Record.NextDomainName);
      for i in RR_Type.WireStringTypeIndex range 1..Current_Name_Length loop
         --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-6)-DNS_Types.message_bytes_range(Current_Name_Length)
         --# and i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last
         --# and Current_Name_Length >= RR_Type.WireStringTypeIndex'First and Current_Name_Length <= RR_Type.WireStringTypeIndex'Last
         --# and (for all J in RR_Type.WireStringTypeIndex => (Character'Pos(NSEC_Record.NextDomainName(J)) >= 0 and (Character'Pos(NSEC_Record.NextDomainName(J)) <= 255)));
         Bytes(((Start_Byte+6)-1)+DNS_Types.message_bytes_range(i)) := DNS_Types.Byte(Character'Pos(NSEC_Record.NextDomainName(i)));
      END LOOP;

      -- RR bitmap section (ugh)
      dstBytes := 0;
      BlockOffset := (((Start_Byte+6)-1)+DNS_Types.message_bytes_range(Current_Name_Length));
      FOR block IN rr_type.nsec_record_type.blockNumberValue range 1..NSEC_Record.NumberOfBlocks LOOP
         --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-6)-DNS_Types.message_bytes_range(Current_Name_Length)
         --# and Current_Name_Length >= RR_Type.WireStringTypeIndex'First and Current_Name_Length <= RR_Type.WireStringTypeIndex'Last
         --# and blockOffset >= dns_types.message_bytes_range'first
         --# and blockOffset <= (((Start_Byte+6)-1)+DNS_Types.message_bytes_range(Current_Name_Length))
         --#  + DNS_Types.message_bytes_range((block-1)*(2+rr_type.nsec_record_type.BlockLengthValue'last))
         --# and dstBytes <= dns_types.message_length_range((block-1)*rr_type.nsec_record_type.BlockLengthValue'last)
         --# and NSEC_Record.BlockLengths(Block) >= rr_type.nsec_record_type.BlockLengthValue'first
         --# and NSEC_Record.BlockLengths(Block) <= rr_type.nsec_record_type.BlockLengthValue'last;
         DstBytes := DstBytes + dns_types.message_length_range(NSEC_Record.BlockLengths(Block)); --sum up the block lengths, needed below
         Bytes(BlockOffset +1) := NSEC_Record.BlockNumbers(Block);
         Bytes(BlockOffset +2) := Dns_Types.Byte(NSEC_Record.BlockLengths(Block));
         FOR Byte IN rr_type.nsec_record_type.BlockLengthValue range 1..NSEC_Record.BlockLengths(Block) LOOP
            --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-6)-DNS_Types.message_bytes_range(Current_Name_Length)
            --# and Current_Name_Length >= RR_Type.WireStringTypeIndex'First and Current_Name_Length <= RR_Type.WireStringTypeIndex'Last
            --# and blockOffset >= dns_types.message_bytes_range'first
            --# and blockOffset <= (((Start_Byte+6)-1)+DNS_Types.message_bytes_range(Current_Name_Length))
            --#  + DNS_Types.message_bytes_range((block-1)*(2+rr_type.nsec_record_type.BlockLengthValue'last))
            --# and dstBytes <= dns_types.message_length_range(block*rr_type.nsec_record_type.BlockLengthValue'last)
            --# and byte >= rr_type.nsec_record_type.BlockLengthValue'first and byte <= rr_type.nsec_record_type.BlockLengthValue'last
            --# and block = block%;
            Bytes((BlockOffset+2)+Dns_Types.message_bytes_range(Byte)) := NSEC_Record.BitMaps(Block)(Byte);
         END LOOP;
         BlockOffset := (BlockOffset + 2) +  dns_types.message_bytes_range(NSEC_Record.BlockLengths(Block));
      END LOOP;
      DstBytes := DstBytes + dns_types.message_length_range(NSEC_Record.NumberOfBlocks*2);   --add in the bytes for the block numbers and the block lengths
      DstBytes := (DstBytes + dns_types.message_length_range(Current_Name_Length));   --sets final value of OUT parameter
      -- RDLENGTH (2 bytes, can now set final value as the length of the domain name plus the length of the bitmapped RRList)
      Set_Unsigned_16(Bytes,Start_Byte+4,unsigned_types.unsigned16(dstBytes));
   END Set_TTL_Data_NSEC_Response;

   procedure Set_TTL_Data_NXDOMAIN_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         NXDOMAIN_Record : in Rr_Type.NXDOMAIN_Record_Type.NXDOMAINRecordType) is
   begin
         -- TTL (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte,NXDOMAIN_Record.ttlInSeconds);
         -- RDLENGTH (2 bytes)
         Set_Unsigned_16(Bytes,Start_Byte+4,0);  --doesn't have any data
   END Set_TTL_Data_NXDOMAIN_Response;

procedure Set_TTL_Data_PTR_Response(
         Bytes               : in out DNS_Types.Bytes_Array_Type;
         Start_Byte          : in DNS_Types.message_bytes_range;
         PTR_Record           : in Rr_Type.ptr_record_type.PTRRecordType;
         Current_Name_Length : in RR_Type.WireStringTypeIndex) is
   begin
         -- TTL
         Set_Unsigned_32(Bytes,Start_Byte,PTR_Record.ttlInSeconds);
         -- DATA # bytes is equal to length of WireString
         Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16(Current_name_Length));
         -- copy NS record
         for i in RR_Type.WireStringTypeIndex range 1..Current_Name_Length loop
            --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-
            --# (6+DNS_Types.message_bytes_range(Current_Name_Length))) and
            --# i >= RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last;
            Bytes((Start_Byte+5)+DNS_Types.message_bytes_range(I)) := DNS_Types.Byte(
               Character'Pos(PTR_Record.domainname(i)));
         end loop;
   end Set_TTL_Data_PTR_Response;

   --see http://www.rfc-editor.org/rfc/rfc4034.txt for RRSIG record format
   procedure Set_TTL_Data_RRSIG_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         RRSIG_Record : IN Rr_Type.RRSIG_Record_Type.RRSIGRecordType;
         dstBytes : OUT DNS_Types.message_bytes_range) IS
         WireVersion : RR_Type.WireStringType;
         Current_Name_Length : RR_Type.WireStringTypeIndex;
         srcByte : Integer;
         sigLength : Rr_Type.Rrsig_Record_Type.sigLengthValueType;
         sigOffset : DNS_Types.message_bytes_range;
         SrcNum0, SrcNum1, SrcNum2, SrcNum3: Dns_Types.Byte;
         DstNum0, DstNum1, DstNum2 : Dns_Types.Byte;
         function From_Query_Type is new Ada.Unchecked_Conversion(DNS_Types.Query_Type,Unsigned_Types.Unsigned32);
         function From_Bytes_Range is new Ada.Unchecked_Conversion(DNS_Types.message_bytes_range,Unsigned_Types.Unsigned32);
   BEGIN
         -- TTL (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte,RRSIG_Record.ttlInSeconds);
         -- RDLENGTH (2 bytes, value is 18 plus the length of the name plus the length of the signature, placeholder at this point)
         Set_Unsigned_16(Bytes,Start_Byte+4,0);
         -- TYPE COVERED (2 bytes)
         --# accept Warning, 12, From_Query_Type, "unchecked conversions ok";
         -- Ugh.  The things we do to make SPARK happy.  mod is superfluous, used only to convince SPARK that the value is in type.
         Set_Unsigned_16(Bytes,Start_Byte+6,Unsigned_Types.Unsigned16(from_query_type(RRSIG_Record.typeCovered) mod 65536));
         -- ALGORITHM (1 byte)
         Bytes(Start_Byte+8) := DNS_Types.Byte(RRSIG_Record.Algorithm);
         -- LABELS (1 byte)
         Bytes(Start_Byte+9) := DNS_Types.Byte(RRSIG_Record.numLabels);
         -- ORIG TTL (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte+10,RRSIG_Record.OrigTTL);
         -- SIG EXPIRATION (4 bytes)
         Set_Unsigned_32(Bytes,Start_Byte+14,RRSIG_Record.SigExpiration);
         -- SIG INCEPTION (4 bytes) (starring Leonardo DiCaprio)
         Set_Unsigned_32(Bytes,Start_Byte+18,RRSIG_Record.SigInception);
         -- KEY TAG (2 bytes)
         Set_Unsigned_16(Bytes,Start_Byte+22,RRSIG_Record.KeyTag);

         -- SIGNER'S NAME (Current_Name_Length bytes)
         WireVersion := Rr_Type.ConvertDomainNameToWire(RRSIG_Record.SignerName);
         Current_Name_Length := RR_Type.WireNameLength(WireVersion);
         for i in RR_Type.WireStringTypeIndex range 1..Current_Name_Length loop
            --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-24)-DNS_Types.message_bytes_range(Current_Name_Length)
            --# and i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last
            --# and Current_Name_Length >= RR_Type.WireStringTypeIndex'First and Current_Name_Length <= RR_Type.WireStringTypeIndex'Last
            --# and (for all J in RR_Type.WireStringTypeIndex => (Character'Pos(WireVersion(J)) >= 0 and (Character'Pos(WireVersion(J)) <= 255)));
            Bytes(((Start_Byte+24)-1)+DNS_Types.message_bytes_range(i)) := DNS_Types.Byte(Character'Pos(WireVersion(i)));
         END LOOP;

         -- SIGNATURE
         srcByte := 1;
         dstBytes := 1;
         sigLength := RRSIG_Record.SignatureLength;
         sigOffset := (((Start_Byte+24)-1)+DNS_Types.message_bytes_range(Current_Name_Length)); --this+1 is the index of the first byte of the signature
         while (srcByte + 3 <= sigLength) loop  --since it's base64 encoded, src key will always contain an exact multiple of 4 bytes
            --# assert srcByte + 3 <= sigLength and
            --# sigLength <= rr_type.RRSIG_Record_Type.MaxRRSIGLength and sigLength = RRSIG_Record.signatureLength and
            --# dstBytes >= DNS_Types.message_bytes_range(1) and dstBytes <= DNS_Types.message_bytes_range(srcByte) and
            --# Start_Byte <= DNS_Types.message_bytes_range'Last-24-DNS_Types.message_bytes_range(rr_type.MaxDomainNameLength)-DNS_Types.message_bytes_range(sigLength)
            --# and Current_Name_Length >= RR_Type.WireStringTypeIndex'First and Current_Name_Length <= RR_Type.WireStringTypeIndex'Last
            --# and sigOffset = Start_Byte+24-1+DNS_Types.message_bytes_range(Current_Name_Length);
            srcNum0 := Base64CharToVal(RRSIG_Record.signature(srcByte));     --00aaaaaa
            srcNum1 := Base64CharToVal(RRSIG_Record.signature(srcByte+1));   --00bbbbbb
            srcNum2 := Base64CharToVal(RRSIG_Record.signature(srcByte+2));   --00cccccc
            SrcNum3 := Base64CharToVal(RRSIG_Record.signature(SrcByte+3));   --00dddddd

            DstNum0 := SrcNum0*4 + (Srcnum1/16);   --aaaaaa00 + 000000bb = aaaaaabb
            DstNum1 := Srcnum1*16 + (SrcNum2/4);   --bbbb0000 + 0000cccc = bbbbcccc
            DstNum2 := SrcNum2*64 + SrcNum3;       --cc000000 + 00dddddd = ccdddddd

            Bytes(sigOffset + dstBytes) := DstNum0;
            Bytes(sigOffset + (dstBytes+1)) := DstNum1;
            Bytes(sigOffset + (dstBytes+2)) := DstNum2;

            srcByte := srcByte + 4;
            dstBytes := dstBytes + 3;
         END LOOP;

         -- Finally, we know how many bytes have been used on the wire (24 plus the length of the signer's name plus the length of the key)
         dstBytes := DNS_Types.message_bytes_range(24+Current_Name_Length)+(dstBytes-DNS_Types.message_bytes_range(1)); --sets the out parameter of this procedure
         -- and we can set RDLENGTH (2 bytes, value is 18 plus the length of the signer's name plus the length of the key)
	   --# accept Warning, 12, From_Bytes_Range, "unchecked conversions ok";
         -- The things we do to make SPARK happy.
         Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16(from_bytes_range(dstBytes-6) mod 65536));
         -- sets RDLENGTH field (TTL and RDLENGTH fields don't count, so subtract 4+2=6 bytes)

   END Set_TTL_Data_RRSIG_Response;

   procedure Set_TTL_Data_SOA_Response(
         Bytes                  : in out DNS_Types.Bytes_Array_Type;
         Start_Byte             : in DNS_Types.message_bytes_range;
         SOA_Record             : in Rr_Type.SOA_record_type.SOARecordType;
         Nameserver_Name_Length : in RR_Type.WireStringTypeIndex;
         Mailbox_Name_Length    : in RR_Type.WireStringTypeIndex) is
      Current_Byte : DNS_Types.message_bytes_range;
   begin
      -- TTL
      Set_Unsigned_32(Bytes,Start_Byte,SOA_Record.ttlInSeconds);
      -- DATA # bytes is equal to length of both WireStrings + 20
      Set_Unsigned_16(Bytes,Start_Byte+4,Unsigned_Types.Unsigned16(
         Nameserver_Name_Length + (Mailbox_Name_Length+20)));
      -- copy NS record
      for i in RR_Type.WireStringTypeIndex range 1..Nameserver_Name_Length loop
         --# assert Start_Byte <= (DNS_Types.message_bytes_range'Last-20)-
         --# DNS_Types.message_bytes_range(Mailbox_Name_Length+Nameserver_Name_Length) and
         --# i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last;
         Bytes((Start_Byte+5)+DNS_Types.message_bytes_range(I)) := DNS_Types.Byte(
            Character'Pos(SOA_Record.nameServer(i)));
      end loop;
      Current_Byte := Start_Byte+DNS_Types.message_bytes_range(
         5+(Nameserver_Name_Length));
      -- copy MB record
      for i in RR_Type.WireStringTypeIndex range 1..Mailbox_Name_Length loop
         --# assert Current_Byte>=1 and Current_Byte <= (DNS_Types.message_bytes_range'Last-20)-
         --# DNS_Types.message_bytes_range(Mailbox_Name_Length) and
         --# i>=RR_Type.WireStringTypeIndex'First and i<= RR_Type.WireStringTypeIndex'Last;
         Bytes(Current_Byte+DNS_Types.message_bytes_range(I)) := DNS_Types.Byte(
            Character'Pos(SOA_Record.email(i)));
      end loop;
      Current_Byte := Start_Byte+DNS_Types.message_bytes_range(
         6+(Nameserver_Name_Length + Mailbox_Name_Length));
      -- serial number
      Set_Unsigned_32(Bytes,Current_Byte,SOA_Record.SerialNumber);
      -- refresh interval
      Set_Unsigned_32(Bytes,Current_Byte+4,SOA_Record.refresh);
      -- retry interval
      Set_Unsigned_32(Bytes,Current_Byte+8,SOA_Record.retry);
      -- expiration limit
      Set_Unsigned_32(Bytes,Current_Byte+12,SOA_Record.expiry);
      -- minimum TTL
      Set_Unsigned_32(Bytes,Current_Byte+16,SOA_Record.minimum);
   end Set_TTL_Data_SOA_Response;

   PROCEDURE Randomize_Case_Of_Query_Name(
      Input_Message  : IN OUT DNS_Types.Dns_Message) IS
      randomVal: non_spark_discrete_random_messageID_pkg.validMessageIDType;
      Byte : Dns_types.message_bytes_range := DNS_Types.message_bytes_range'First;
      I : Natural := RR_Type.WireStringType'First;

      type bit is mod 2;
      SUBTYPE Bit_Number_for_Short_Type IS Natural RANGE 0..15;
      --MSB is 15, LSB is 0
      FUNCTION Bit_Number(Index : IN Bit_Number_for_Short_Type;
         val : in dns_types.Unsigned_Short) return bit is
         retval : bit;
      BEGIN
         IF (Val/(2**index)) mod 2 = 0 then
            retval := 0;
         ELSE
            retval := 1;
         END IF;
         return retval;
      END Bit_Number;

   BEGIN
      randomVal := non_spark_discrete_random_messageID_pkg.RandomID;
      while Byte < DNS_Types.message_bytes_range'Last-4 and then Input_message.Bytes(Byte)/=0
         and then I < RR_Type.WireStringType'Last loop
         --# assert I>=RR_Type.WireStringType'First and I < RR_Type.WireStringType'Last and
         --# Byte >= DNS_Types.message_bytes_range'First
         --# and Byte < DNS_Types.message_bytes_range'Last-4;

         --this is in wire format, so skip over the bytes that are
         --length fields
         IF Ada.Characters.Handling.Is_Alphanumeric(
         Character'Val(Input_Message.Bytes(Byte))) THEN
            IF Bit_Number(I mod 16, randomVal) = 0 THEN
               Input_Message.Bytes(Byte) := Dns_Types.Byte(Character'Pos(
                  Ada.Characters.Handling.To_Lower(Character'Val(Input_Message.Bytes(Byte)))));
            ELSE
               Input_Message.Bytes(Byte) := Dns_Types.Byte(Character'Pos(
                  Ada.Characters.Handling.To_Upper(Character'Val(Input_Message.Bytes(Byte)))));
            END IF;
         end if;
         I := I + 1;
         --get 16 more random bits if you've used up the current batch
         IF (I mod 16) = 0 THEN
            RandomVal := Non_Spark_Discrete_Random_MessageID_Pkg.RandomID;
         end if;
         Byte := Byte + 1;
      end loop;
   END Randomize_Case_Of_Query_Name;

   --should only be called when the domain name argument is the same
   --length as the name it is replacing in the message
   PROCEDURE Set_Query_Name(
      Input_Message  : IN OUT DNS_Types.Dns_Message;
      DomainName : IN RR_Type.WireStringType)
   IS
      Byte : Dns_types.message_bytes_range;
      I : Natural;
   BEGIN
      Byte := DNS_Types.message_bytes_range'First;
      I:= RR_Type.WireStringType'First;
      WHILE Byte < DNS_Types.Message_Bytes_Range'Last-4 AND THEN
         Input_message.Bytes(Byte)/=0 and then I < RR_Type.WireStringType'Last loop
         --# assert I>=RR_Type.WireStringType'First and I < RR_Type.WireStringType'Last and
         --# Byte >= DNS_Types.message_bytes_range'First
         --# and Byte < DNS_Types.message_bytes_range'Last-4;
         --Domainname(I) := Character'Val(Input_Message.Bytes(Byte));
         Input_Message.Bytes(Byte) := Character'Pos(DomainName(I));
         I := I + 1;
         Byte := Byte + 1;
      end loop;
      Input_Message.Bytes(Byte):= 0;  --just in case
   END Set_Query_Name;

   function Get_Query_Name(
      Input_message  : in DNS_Types.dns_message) return RR_Type.WireStringType
   IS
      Byte : Dns_types.message_bytes_range;
      I : Natural;
      Domainname : RR_Type.WireStringType;
   BEGIN
      Byte := DNS_Types.message_bytes_range'First;
      I:= RR_Type.WireStringType'First;
      Domainname := RR_Type.WireStringType'(others => ' ');
      WHILE Byte < DNS_Types.Message_Bytes_Range'Last-4 AND THEN
         Input_message.Bytes(Byte)/=0 and then I < RR_Type.WireStringType'Last loop
         --# assert I>=RR_Type.WireStringType'First and I < RR_Type.WireStringType'Last and
         --# Byte >= DNS_Types.message_bytes_range'First
         --# and Byte < DNS_Types.message_bytes_range'Last-4;
         Domainname(I) := Character'Val(Input_message.Bytes(Byte));
         I := I + 1;
         Byte := Byte + 1;
      end loop;
      Domainname(I) := Character'Val(0);
      return Domainname;
   END Get_Query_Name;

   procedure Get_Query_Name_Type_Class(
         Input_message  : in DNS_Types.dns_message;
         Domainname    : out RR_Type.wireStringType;
         Query_Type    : out DNS_Types.Query_Type;
         Query_Class   : out DNS_Types.Query_Class;
         End_Byte      : out Dns_types.message_bytes_range) is
      Byte : Dns_types.message_bytes_range := DNS_Types.message_bytes_range'First;
      I : Natural := RR_Type.WireStringType'First;
      QT_Natural, QC_Natural : Natural;
      function Type_To_Natural is new Ada.Unchecked_Conversion(Dns_Types.Query_Type,Natural);
      function to_type is new ada.Unchecked_Conversion(natural,dns_types.query_type);
      function class_to_natural is new ada.Unchecked_Conversion(dns_types.query_class,natural);
      function To_Class is new Ada.Unchecked_Conversion(Natural,Dns_Types.Query_Class);
   begin
      Domainname := RR_Type.WireStringType'(others => ' ');
      while Byte < DNS_Types.message_bytes_range'Last-4 and then Input_message.Bytes(Byte)/=0
         and then I < RR_Type.WireStringType'Last loop
         --# assert I>=RR_Type.WireStringType'First and I < RR_Type.WireStringType'Last and
         --# Byte >= DNS_Types.message_bytes_range'First
         --# and Byte < DNS_Types.message_bytes_range'Last-4;
         Domainname(I) := Character'Val(Input_message.Bytes(Byte));
         I := I + 1;
         Byte := Byte + 1;
      end loop;
      Domainname(I) := Character'Val(0);
--      for I in Byte-2..Byte+4 loop
--         Ada.Text_IO.Put_Line("byte: " & Dns_Types.message_bytes_range'Image(I) & ":" &
--            Natural'Image(natural(Input_message.Bytes(I))));
--      end loop;
      QT_Natural := Natural(Input_message.Bytes(Byte+1))*256+Natural(Input_message.Bytes(Byte+2));
      QC_Natural := Natural(Input_message.Bytes(Byte+3))*256+Natural(Input_message.Bytes(Byte+4));
--      ada.Text_IO.put_line("qt: " & natural'image(qt_natural));
--      ada.Text_IO.put_line("qc: " & natural'image(qc_natural));
      --# accept Warning, 12, type_To_Natural, "unchecked conversions ok";
      if QT_Natural >= type_To_Natural(DNS_Types.Query_Type'First) and QT_Natural <= type_To_Natural(DNS_Types.Query_Type'Last) then
      --# end accept;
         --# accept Warning, 12, To_Type, "unchecked conversions ok";
         Query_Type := To_Type(QT_Natural);
         --# end accept;
         if not Query_Type'Valid then
            Query_Type := DNS_Types.UNIMPLEMENTED;
         end if;
      else
         Query_Type := DNS_Types.UNIMPLEMENTED;
      END IF;
      IF Query_Type = DNS_Types.UNIMPLEMENTED THEN
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            "unimplemented query type: ", 0);
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            QT_Natural,2,10);
          Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
      END IF;

      --# accept Warning, 12, class_To_Natural, "unchecked conversions ok";
      if QC_Natural >= class_to_natural(DNS_Types.Query_Class'First) and QC_Natural <= class_to_natural(DNS_Types.Query_Class'Last) then
      --# end accept;
         --# accept Warning, 12, To_Class, "unchecked conversions ok";
         Query_Class := To_Class(QC_Natural);
         --# end accept;
         if not Query_Class'Valid then
            Query_Class := DNS_Types.NONE_CLASS;
         end if;
      else
         Query_Class := DNS_Types.NONE_CLASS;
      end if;
      IF Query_Class = DNS_Types.NONE_CLASS THEN
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            "invalid query class: ", 0);
         Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            QC_Natural,2,10);
          Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
            Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
      END IF;
      End_Byte := (Byte + 4);
   end Get_Query_Name_Type_Class;

   procedure Create_Response_Error(
         Input_Bytes   : in DNS_Types.message_length_range;
         Output_message : in out DNS_Types.dns_message;
         Output_Bytes  : out DNS_Types.message_length_range) is
   begin
      Output_message.Header.AA := True;
      Output_message.Header.RCODE   := DNS_Types.Not_Implemented;
      Output_message.Header.ANCount := 0;
      Output_Bytes := Input_Bytes;
   end Create_Response_Error;


   procedure Create_Response_Zero(
         Start_Byte    : in DNS_Types.Message_Bytes_Range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         output_message : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes  : out DNS_Types.message_Length_Range) IS SEPARATE;

   procedure Create_Response_A(
         Start_Byte    : in DNS_Types.message_bytes_range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
      Output_Bytes  : OUT DNS_Types.Message_Length_Range) IS SEPARATE;

   procedure Create_Response_AAAA(
         Start_Byte    : in DNS_Types.message_bytes_range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes  : out DNS_Types.message_length_range) is separate;

   procedure Create_Response_MX(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : IN DNS_Types.QNAME_PTR_RANGE;
         output_message  : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : OUT DNS_Types.Message_Length_Range;
         Num_Found       : out RR_Type.NumberOfRecordsType;
         Qname_Locations : out QNAME_PTR_RANGE_Array;
         Replies         : out RR_Type.mx_record_type.MXRecordBucketType) IS SEPARATE;

   procedure Create_Response_NS(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : IN DNS_Types.QNAME_PTR_RANGE;
         output_message  : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : OUT DNS_Types.Message_Length_Range;
         Num_Found       : out RR_Type.NumberOfRecordsType;
         Qname_Locations : out QNAME_PTR_RANGE_Array;
      Replies         : OUT RR_Type.Ns_Record_Type.NSRecordBucketType) IS SEPARATE;

   procedure Create_Response_PTR(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : in DNS_Types.QNAME_PTR_RANGE;
         Output_message   : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : OUT DNS_Types.message_length_range) IS SEPARATE;

   procedure Create_Response_SOA(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : in DNS_Types.QNAME_PTR_RANGE;
         Output_message   : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : out DNS_Types.message_length_range) is separate;

   procedure Create_Response_DNSKEY(
         Start_Byte    : in DNS_Types.message_bytes_range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
      Output_Bytes  : OUT DNS_Types.message_length_range) IS SEPARATE;

   procedure Create_Response_NSEC(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
      Output_Bytes   : OUT DNS_Types.Message_Length_Range) IS SEPARATE;

   procedure Create_Response_RRSIG(
         Start_Byte    : in DNS_Types.message_bytes_range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
      Output_Bytes  : OUT DNS_Types.message_length_range) IS SEPARATE;

   procedure Create_Response_NXDOMAIN(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : OUT DNS_Types.message_length_range) IS separate;

   procedure Create_Response_EDNS(
         Input_message      : in DNS_Types.dns_message;
         Input_Bytes        : in DNS_Types.message_length_range;
         Query_End_Byte     : in DNS_Types.message_bytes_range;
         Start_Byte         : in DNS_Types.message_bytes_range;
         Output_message     : in out DNS_Types.dns_message;
         Output_Bytes       : out DNS_Types.message_length_range;
         Additional_Count   : in out DNS_Types.Unsigned_Short;
         DNSSEC             : out Boolean;
         Max_Transmit       : out DNS_Types.message_length_range) is separate;

      --this only checks the first entry in Cnames.  If Cnames has > 1
   --entry, a warning message will have been previously printed.
   procedure Process_Response_Cname(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Cnames         : in RR_Type.cname_record_type.CNAMERecordBucketType;
         Domainname     : OUT RR_Type.WireStringType;
         Qname_Location : in out DNS_Types.QNAME_PTR_RANGE;
         Output_Message  : IN OUT DNS_Types.Dns_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : OUT DNS_Types.Message_Length_Range) IS SEPARATE;

   procedure Create_Response(
         Input_message  : in DNS_Types.dns_message;
         Input_Bytes   : in DNS_Types.message_length_range;
         Output_message : in out DNS_Types.dns_message;
         Output_Bytes  : out DNS_Types.message_length_range;
         Max_Transmit  : OUT DNS_Types.Message_Length_Range;
         Source_Address : DNS_Network.Network_Address_And_Port;
         Domainname_Out : OUT RR_Type.WireStringType) IS SEPARATE;

 end Process_Dns_Request;

