----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--WITH Ada.Text_IO;
WITH DNS_Types;
with DNS_Table_Pkg;
with DNS_Network;
with RR_Type;
with RR_Type.a_record_type;
with RR_Type.Aaaa_Record_Type;
WITH RR_Type.Cname_Record_Type;
with RR_Type.DNSKEY_Record_Type;
with RR_Type.ns_record_type;
WITH RR_Type.Mx_Record_Type;
WITH RR_Type.NSEC_Record_Type;
with RR_Type.NXDOMAIN_Record_Type;
WITH RR_Type.Ptr_Record_Type;
with Rr_Type.RRSIG_Record_Type;
with RR_Type.soa_record_type;
WITH Unsigned_Types;
WITH Non_Spark_Discrete_Random_MessageID_Pkg;
WITH Ada.Characters.Handling;
WITH Ada.Unchecked_Conversion;
WITH Protected_SPARK_IO_05;

With Ada.Strings.fixed;
With Ada.Strings.Maps;
With Ada.Strings.Maps.constants;
With ADA.Text_IO;

--# inherit DNS_types, DNS_Network, DNS_Network_Receive, System, Protected_SPARK_IO_05,
--#          RR_Type, RR_Type.a_record_type, RR_Type.aaaa_record_type,
--#          RR_Type.cname_record_type, RR_Type.dnskey_record_type, RR_type.mx_record_type,
--#          RR_Type.ns_record_type, RR_Type.NSEC_Record_Type, RR_Type.NXDOMAIN_Record_Type,
--#          RR_Type.ptr_record_type, Rr_Type.RRSIG_Record_Type, RR_type.soa_record_type,
--#          DNS_Table_Pkg, Unsigned_Types, ada.Unchecked_Conversion, ada.real_time,
--#	     ada.characters.handling, non_spark_discrete_random_messageID_pkg;

package Process_Dns_Request is
  procedure Create_Response(
         Input_message  : in DNS_Types.dns_message;
         Input_Bytes   : in DNS_Types.Message_Length_Range;
         Output_message : in out DNS_Types.dns_message;
         Output_Bytes  : out DNS_Types.Message_Length_Range;
         Max_Transmit  : out DNS_Types.Message_Length_Range;
         Source_Address : in DNS_Network.Network_Address_And_Port;
         Domainname_Out : out RR_Type.WireStringType);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --#        in out Protected_SPARK_IO_05.SPARK_IO_PO;
   --# derives Output_message from *, DNS_Table_Pkg.DNS_Table, Input_message,
   --#          Ada.Real_Time.ClockTime, Input_Bytes &
   --#    Output_Bytes from Input_Bytes, Input_message,
   --#       DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Protected_SPARK_IO_05.SPARK_IO_PO from *, DNS_Table_Pkg.DNS_Table,
   --#       Input_message, Source_address, Ada.Real_Time.ClockTime &
   --#    Domainname_Out from Input_Message,
   --#       DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Max_Transmit from DNS_Table_Pkg.DNS_Table, Input_message, Input_Bytes,
   --#       Ada.Real_Time.ClockTime;
   --# pre Integer(Input_Bytes) >=DNS_Types.Header_Bits/8+1;
   --     and Integer(Input_Bytes) < 312; --need to modify this condition to get DNSSEC to validate
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and
   --#    Integer(Output_Bytes) <= DNS_Types.message_size and
   --#    Integer(Max_Transmit) <= DNS_Types.Message_Size and
   --#    Max_Transmit >= DNS_Types.UDP_Max_Size;

   PROCEDURE Randomize_Case_Of_Query_Name(
      Input_Message  : IN OUT DNS_Types.Dns_Message);
   --# derives Input_Message from Input_Message;

   PROCEDURE Set_Query_Name(
      Input_Message  : IN OUT DNS_Types.Dns_Message;
      DomainName : in RR_Type.WireStringType);
   --# derives Input_Message from Input_Message, DomainName;

   function Get_Query_Name(
      Input_message  : in DNS_Types.dns_message) return RR_Type.WireStringType;

   procedure Get_Query_Name_Type_Class(
         Input_message  : in DNS_Types.dns_message;
         Domainname    : out RR_Type.wireStringType;
         Query_Type    : out DNS_Types.Query_Type;
         Query_Class   : out DNS_Types.Query_Class;
         End_Byte      : out Dns_types.message_bytes_range);
   --# global in out Protected_SPARK_IO_05.SPARK_IO_PO;
   --# derives Domainname from Input_message &
   --#         Query_Type from Input_message &
   --#         Query_Class from Input_message &
   --#         End_Byte from Input_message &
   --# 	       Protected_SPARK_IO_05.SPARK_IO_PO from Input_Message, Protected_SPARK_IO_05.SPARK_IO_PO;
   --# post End_Byte >= 4;
private
   procedure Set_Unsigned_32(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.Message_Bytes_Range;
         Value : in Unsigned_Types.Unsigned32);
   --# derives Bytes from *, Start_Byte, Value;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-3;

   procedure Set_Unsigned_16(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.Message_Bytes_Range;
         Value : in Unsigned_Types.Unsigned16);
   --# derives Bytes from *, Start_Byte, Value;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-1;

   procedure Set_TTL_Data_A_Response(
      Bytes      : in out DNS_Types.Bytes_Array_Type;
      Start_Byte : in DNS_Types.message_bytes_range;
      A_Record   : in Rr_Type.A_Record_Type.ARecordType);
   --# derives Bytes from *, Start_Byte, A_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-10;
   -- (10 because for A record TTL + RDLENGTH + RDATA field = 10 bytes)


   procedure Set_TTL_Data_AAAA_Response(
      Bytes      : in out DNS_Types.Bytes_Array_Type;
      Start_Byte : in DNS_Types.message_bytes_range;
      AAAA_Record   : in Rr_Type.AAAA_Record_Type.AAAARecordType);
   --# derives Bytes from *, Start_Byte, AAAA_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-22;
   -- (22 because for AAAA record TTL + RDLENGTH + RDATA field = 22 bytes)

   procedure Set_TTL_Data_DNSKEY_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         DNSKEY_Record : IN Rr_Type.DNSKEY_Record_Type.DNSKEYRecordType;
         dstBytes : out rr_type.dnskey_record_type.keyLengthValueType);
   --# derives Bytes from *, Start_Byte, DNSKEY_Record &
   --#	dstBytes from DNSKEY_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last -
   --#   (10+DNS_Types.message_bytes_range(DNSKEY_Record.keyLength));
   -- (DNSKEY record from TTL field is at most 10 + key length bytes long)

   procedure Set_TTL_Data_NSEC_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         NSEC_Record : IN Rr_Type.NSEC_Record_Type.NSECRecordType;
         dstbytes : OUT DNS_Types.message_length_range);
   --# derives Bytes from *, Start_Byte, NSEC_Record & dstBytes from NSEC_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last -
   --#    (6+DNS_Types.message_bytes_range(rr_type.maxDomainNameLength+1)+
   --#    DNS_Types.message_bytes_range(rr_type.nsec_record_type.maxRRDataLength));
   --  (NSEC record from TTL field is at most 6 + max name length +
   --  max resource record length bytes long)
   --# post dstBytes <= DNS_Types.message_length_range(Rr_Type.NSEC_Record_Type.MaxRRDataLength);

   procedure Set_TTL_Data_NXDOMAIN_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         NXDOMAIN_Record : IN Rr_Type.NXDOMAIN_Record_Type.NXDOMAINRecordType);
   --# derives Bytes from *, Start_Byte, NXDOMAIN_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-6+1;
   -- (6-1 because for NXDOMAIN record TTL + RDLENGTH = 6 bytes,
   -- (no RDATA field), +1 because Start_Byte is first byte used)

   procedure Set_TTL_Data_RRSIG_Response(
         Bytes : in out DNS_Types.Bytes_Array_Type;
         Start_Byte : in DNS_Types.message_bytes_range;
         RRSIG_Record : IN Rr_Type.RRSIG_Record_Type.RRSIGRecordType;
         dstbytes : OUT DNS_Types.message_bytes_range);
   --# derives Bytes from *, Start_Byte, RRSIG_Record &
   --#	dstBytes from RRSIG_Record;
   --# pre Start_Byte <= DNS_Types.message_bytes_range'Last-24-DNS_Types.message_bytes_range(Rr_Type.MaxDomainNameLength) -
   --#	DNS_Types.message_bytes_range(Rr_Type.Rrsig_Record_Type.maxrrsigLength);
   --# post dstBytes <= DNS_Types.message_bytes_range(24 + Rr_Type.MaxDomainnameLength + Rr_Type.RRSIG_Record_Type.maxrrsigLength);

   procedure Set_TTL_Data_NS_Response(
         Bytes               : in out DNS_Types.Bytes_Array_Type;
         Start_Byte          : in DNS_Types.message_bytes_range;
         NS_Record           : in Rr_Type.ns_record_type.NSRecordType;
         Current_Name_Length : in RR_Type.WireStringTypeIndex);
   --# derives Bytes from *, Start_Byte, NS_Record, Current_Name_Length;
   --# pre Current_Name_Length >= 0 and Current_Name_Length<=RR_Type.WireStringTypeIndex'Last and
   --#     Start_Byte <= (DNS_Types.message_bytes_range'Last -
   --#    (6+DNS_Types.message_bytes_range(Current_Name_Length)));

   procedure Set_TTL_Data_PTR_Response(
         Bytes                : in out DNS_Types.Bytes_Array_Type;
         Start_Byte           : in DNS_Types.message_bytes_range;
         PTR_Record           : in Rr_Type.ptr_record_type.PTRRecordType;
         Current_Name_Length  : in RR_Type.WireStringTypeIndex);
   --# derives Bytes from *, Start_Byte, PTR_Record, Current_Name_Length;
   --# pre Current_Name_Length >= 0 and Current_Name_Length<=RR_Type.WireStringTypeIndex'Last and
   --#     Start_Byte <= (DNS_Types.message_bytes_range'Last -
   --#    (6+DNS_Types.message_bytes_range(Current_Name_Length)));

   procedure Set_TTL_Data_MX_Response(
         Bytes               : in out DNS_Types.Bytes_Array_Type;
         Start_Byte          : in DNS_Types.message_bytes_range;
         MX_Record           : in Rr_Type.MX_record_type.MXRecordType;
         Current_Name_Length : in RR_Type.WireStringTypeIndex);
   --# derives Bytes from *, Start_Byte, MX_Record, Current_Name_Length;
   --# pre Current_Name_Length >= 0 and Current_Name_Length<=RR_Type.WireStringTypeIndex'Last and
   --#     Start_Byte <= (DNS_Types.message_bytes_range'Last -
   --#    (8+DNS_Types.message_bytes_range(Current_Name_Length)));

   procedure Set_TTL_Data_SOA_Response(
         Bytes                  : in out DNS_Types.Bytes_Array_Type;
         Start_Byte             : in DNS_Types.message_bytes_range;
         SOA_Record             : in Rr_Type.SOA_record_type.SOARecordType;
         Nameserver_Name_Length : in RR_Type.WireStringTypeIndex;
         Mailbox_Name_Length    : in RR_Type.WireStringTypeIndex);
   --# derives Bytes from *, Start_Byte, SOA_Record, Nameserver_Name_Length, Mailbox_Name_Length;
   --# pre Nameserver_Name_Length >= 0 and Nameserver_Name_Length<=RR_Type.WireStringTypeIndex'Last and
   --#     Mailbox_Name_Length >= 0 and Mailbox_Name_Length<=RR_Type.WireStringTypeIndex'Last and
   --#     Start_Byte <= (DNS_Types.message_bytes_range'Last -
   --#     (26+DNS_Types.message_bytes_range(Nameserver_Name_Length+Mailbox_Name_Length)));

   procedure Create_Response_EDNS(
         Input_message       : in DNS_Types.dns_message;
         Input_Bytes        : in DNS_Types.message_length_range;
         Query_End_Byte     : in DNS_Types.message_bytes_range;
         Start_Byte         : in DNS_Types.message_bytes_range;
         Output_message      : in out DNS_Types.dns_message;
         Output_Bytes       : out DNS_Types.message_length_range;
         Additional_Count   : in out DNS_Types.Unsigned_Short;
         DNSSEC             : out Boolean;
         Max_Transmit       : out DNS_Types.message_length_range);
   --# derives Output_message from *, Start_Byte, Input_message, Input_Bytes, Query_End_Byte &
   --#         Output_Bytes from Start_Byte, Input_message, Input_Bytes, Query_End_Byte &
   --#         Max_Transmit from Start_Byte, Input_Bytes, Input_message, Query_End_Byte &
   --#         DNSSEC from Start_Byte, Input_Bytes, Input_message, Query_End_Byte &
   --#         Additional_Count from *, Input_message, Start_Byte, Input_Bytes, Query_End_Byte;
   --# pre Additional_Count < DNS_Types.Unsigned_Short'Last;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Output_Bytes <= DNS_Types.Message_Size and
   --#      Additional_Count >= Additional_Count~ and Additional_Count <= Additional_Count~+1 and
   --#      Max_Transmit >= DNS_Types.UDP_Max_Size and Max_Transmit <= DNS_Types.Message_Size;


   procedure Create_Response_Zero(
         Start_Byte    : in DNS_Types.Message_Bytes_Range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         output_message : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes  : out DNS_Types.message_Length_Range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Qname_Location, Start_Byte, Domainname,
   --#       DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords)
   --#     and Integer(Start_Byte) <= DNS_Types.Message_Size;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);


   procedure Create_Response_A(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Qname_Location, Start_Byte, Domainname,
   --#       DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords)
   --#     and Integer(Start_Byte) <= DNS_Types.Message_Size;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

procedure Create_Response_AAAA(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size and
   --#     Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~ + DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   procedure Create_Response_DNSKEY(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords)
   --#     and Integer(Start_Byte) <= DNS_Types.Message_Size;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   procedure Create_Response_NSEC(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   procedure Create_Response_NXDOMAIN(
         Start_Byte    : in DNS_Types.Message_Bytes_Range;
         Domainname    : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         output_message : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes  : OUT DNS_Types.Message_Length_Range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Start_Byte &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;

   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords)
   --#    and Start_Byte + 12 <= DNS_Types.message_bytes_range'Last;
   --  (Need enough room to write 12 bytes of data)


   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);


   procedure Create_Response_RRSIG(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Domainname     : in RR_Type.WireStringType;
         Qname_Location : in DNS_Types.QNAME_PTR_RANGE;
         Output_message  : in out DNS_Types.dns_message;
         Answer_Count   : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.ClockTime;
   --# pre Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords)
   --#     and Integer(Start_Byte) <= DNS_Types.Message_Size;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#      Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   --maximum number of CNAME records in an alias chain to follow
   MaxCNAMEChainLength : constant dns_types.Unsigned_Short := 4;
   procedure Process_Response_Cname(
         Start_Byte     : in DNS_Types.message_bytes_range;
         Cnames         : in RR_Type.cname_record_type.CNAMERecordBucketType;
         Domainname     : out RR_Type.WireStringType;
         Qname_Location : in out DNS_Types.QNAME_PTR_RANGE;
         Output_Message  : IN OUT DNS_Types.Dns_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes   : out DNS_Types.message_length_range);
   --# derives Domainname from Cnames &
   --#         Qname_Location from Start_Byte &
   --#         Output_message from *, Qname_Location, Start_Byte, Cnames &
   --#         Answer_Count from Answer_Count &
   --#         Output_Bytes from Cnames, Start_Byte;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size;
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and
   --#    Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#    ((Answer_Count = Answer_Count~ + 1) or
   --# 	  Answer_Count = dns_types.unsigned_short'last);

   procedure Create_Response_Error(
         Input_Bytes   : in DNS_Types.message_length_range;
         Output_message : in out DNS_Types.dns_message;
         Output_Bytes  : out DNS_Types.message_length_range);
   --# derives Output_message from * &
   --#         Output_Bytes from Input_Bytes;
   --# post Output_Bytes = Input_Bytes;

   type QNAME_PTR_RANGE_Array is array(RR_Type.ReturnedRecordsIndexType) of DNS_Types.QNAME_PTR_RANGE;

   procedure Create_Response_NS(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : IN DNS_Types.QNAME_PTR_RANGE;
         output_message  : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : OUT DNS_Types.Message_Length_Range;
         Num_Found       : out RR_Type.NumberOfRecordsType;
         Qname_Locations : out QNAME_PTR_RANGE_Array;
         Replies         : out RR_Type.ns_record_type.NSRecordBucketType);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.Clocktime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.Clocktime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table,
   --#       Ada.Real_Time.Clocktime &
   --#    Num_Found from DNS_Table_Pkg.DNS_Table, Domainname, Ada.Real_Time.ClockTime &
   --#    Replies from DNS_Table_Pkg.DNS_Table, Domainname, Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Qname_Locations from Start_Byte, DNS_Table_Pkg.DNS_Table, Domainname,
   --#       Ada.Real_Time.ClockTime;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size and
   --#     Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#     Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   procedure Create_Response_PTR(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : in DNS_Types.QNAME_PTR_RANGE;
         Output_message   : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#         Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#         Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size and
   --#     Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#     Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);
   procedure Create_Response_MX(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : IN DNS_Types.QNAME_PTR_RANGE;
         output_message  : in out DNS_Types.DNS_Message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : OUT DNS_Types.Message_Length_Range;
         Num_Found       : out RR_Type.NumberOfRecordsType;
         Qname_Locations : out QNAME_PTR_RANGE_Array;
         Replies         : out RR_Type.mx_record_type.MXRecordBucketType);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#         Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#         Num_Found from DNS_Table_Pkg.DNS_Table, Domainname, Ada.Real_Time.ClockTime &
   --#         Replies from DNS_Table_Pkg.DNS_Table, Domainname, Ada.Real_Time.ClockTime &
   --#         Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#         Qname_Locations from Start_Byte, DNS_Table_Pkg.DNS_Table, Domainname, Ada.Real_Time.ClockTime ;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size and
   --#     Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#     Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

   procedure Create_Response_SOA(
         Start_Byte      : in DNS_Types.message_bytes_range;
         Domainname      : in RR_Type.WireStringType;
         Qname_Location  : in DNS_Types.QNAME_PTR_RANGE;
         Output_message   : in out DNS_Types.dns_message;
         Answer_Count    : in out DNS_Types.Unsigned_Short;
         Output_Bytes    : out DNS_Types.message_length_range);
   --# global in DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# derives Output_message from *, Start_Byte, Qname_Location, Domainname,
   --#    DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Output_Bytes from Domainname, Start_Byte, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime &
   --#    Answer_Count from *, Domainname, DNS_Table_Pkg.DNS_Table, Ada.Real_Time.ClockTime;
   --# pre Integer(Start_Byte) <= DNS_Types.Message_Size and
   --#     Answer_Count <= DNS_Types.Unsigned_Short'Last-DNS_types.Unsigned_Short(rr_type.MaxNumRecords);
   --# post Integer(Output_Bytes) >= DNS_Types.Header_Bits/8+1 and Integer(Output_Bytes) <= DNS_Types.Message_Size and
   --#     Answer_Count <= Answer_Count~+DNS_Types.Unsigned_Short(rr_type.MaxNumRecords);

end Process_Dns_Request;

