----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--with ada.Text_IO;
WITH Non_Spark_Command_Line_Pkg;
with dns_types;
use type dns_types.Response_Code;
with check_patterns;
package body Udp_query_task_pkg

is

   Query_Task : Udp_Query_Task_Type;  --THIS IS THE QUERY TASK
   -----------------------
   -- Udp_query_Task -----
   -----------------------

   --uses global udp_socket declared in dns_network.adb
   TASK BODY Udp_query_Task_type

   IS


      Input_Message : DNS_Types.Dns_Message;
      Input_Bytes   : DNS_Types.message_length_range;
      Output_Message : DNS_Types.dns_message;
      Output_Bytes  : DNS_Types.message_length_range;
      Max_Transmit  : DNS_Types.message_length_range;
      Source_Address : DNS_Network.Network_Address_And_Port;
      Upstream_Server : DNS_Network.Network_Address_And_Port;
      QueryBufferFull : Boolean;
      ResponseBufferFull : boolean;
      Failure : Boolean;
      Timeout : Boolean;
      NumMsgsInQueryBuffer : buffer_pkg.BufferIndexWithZeroType;
      WhereAdded : Buffer_Pkg.BufferIndexType;
      PortNumber : Integer;
      Domainname : RR_Type.WireStringType;


      --consolidated code into a single procedure for readability
      --takes the input message and 1) assigns it a new id, 2) connects
      --the original query to it, 3) preserves the original casing
      --of the domain name, and 4) assigns a randomized case version to
      --the new query
      PROCEDURE PrepareUpstreamQuery(
         SlotOfOriginalQuery : IN Buffer_Pkg.BufferIndexType;
         Input_Message : IN OUT DNS_Types.DNS_Message)
      --# global in out global_buffers.QueryBuffer;
      --# derives global_buffers.QueryBuffer from *, Input_Message,
      --#    SlotOfOriginalQuery &
      --#    Input_Message from *;
      IS
         NewID : Non_Spark_Discrete_Random_MessageID_Pkg.ValidMessageIDType;
         OriginalCaseDomainName : RR_Type.WireStringType;
         RandomizedCaseDomainName : RR_Type.WireStringType;
      BEGIN
          --generate a random ID for the corresponding query to the upstream server
          --(defense against cache poisoning)
          NewID := Non_Spark_Discrete_Random_MessageID_Pkg.RandomID;

          --connect the two through the ConnectingID field
          Global_Buffers.QueryBuffer.SetConnectingId(SlotOfOriginalQuery , NewID);
          Input_Message.Header.MessageID := NewId;

          --preserve casing of domain name in original query
          --for when we need to send the response
          OriginalCaseDomainName :=
             Process_Dns_Request.Get_Query_Name(Input_Message);
          Global_Buffers.QueryBuffer.SetOriginalCaseDomainName(
             SlotOfOriginalQuery, OriginalCaseDomainName);

          --randomize the case of each letter of the query name as
          --a further defense against cache poisoning
          Process_Dns_Request.Randomize_Case_Of_Query_Name(Input_Message);
          RandomizedCaseDomainName :=
             Process_Dns_Request.Get_Query_Name(Input_Message);
          Global_Buffers.QueryBuffer.SetRandomizedCaseDomainName(
             SlotOfOriginalQuery, RandomizedCaseDomainName);
      END PrepareUpstreamQuery;
   BEGIN
      non_spark_command_line_pkg.Process_Command_Line;

      --replaced Initialize_UDP
      Dns_Network.Initialize_UDP_State;

      --set the upstream server we're going to use
      Dns_Network.SetAddrAndPortCommandLine(Upstream_Server,
         Dns_Network.DNS_PORT);

      --initialize the buffer names
      global_buffers.QueryBuffer.SetName   ("Query buffer    ");
      global_buffers.ResponseBuffer.SetName("Response buffer ");

      --can now start the response task
      Udp_Response_Task_Pkg.StartResponseTask;

      loop
         --# assert true;
         --wait here until a DNS message arrives on either the dedicated
         --downstream socket or one of the ephemeral upstream sockets
         DNS_Network_Receive.Receive_DNS_Message(
            Message       => Input_Message,
            Number_Bytes  => Input_Bytes,
            Reply_Address => Source_Address,
            Arriving_Port => PortNumber,
            Failure       => Failure,
            Timeout       => Timeout);
         Global_Buffers.QueryBuffer.CheckBufferFull(QueryBufferFull);
         Global_Buffers.ResponseBuffer.CheckBufferFull(ResponseBufferFull);
         NumMsgsInQueryBuffer := Global_Buffers.QueryBuffer.NumMsgsInBuffer;

         --Ada.Text_IO.Put("Got a message, Input_Bytes = " &
            --dns_types.message_length_range'Image(Input_Bytes));
         --Ada.Text_IO.New_Line;

         IF (Failure OR Integer(Input_Bytes) <= DNS_Types.Header_Bits/8) AND
         not Timeout THEN
            Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
               Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
               "Receive of query from downstream server failed",0);
         ELSIF Timeout and NumMsgsInQueryBuffer > 0 THEN
            --Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
            --   Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            --   "Waiting for upstream DNS responses (",0);
            --Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
            --   Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            --   Integer(NumMsgsInQueryBuffer),1,10);
            --Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
            --   Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
            --   " messages in query buffer) ",0);
            --delay 0.00001;
            null;
         ELSIF Timeout THEN
            --Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,"Waiting for downstream DNS queries ...",0);
	    --delay 0.00001;
            null;
            --valid message received, is it query or response?
         ELSIF Dns_Types.Is_Query(Input_Message.Header) THEN --know we have a valid query
            if not QueryBufferFull then
               Output_Message := dns_types.EMPTY_MSG;
               --this checks the DNS table/cache
               Process_Dns_Request.Create_Response(
                  Input_Message => Input_Message,
                  Input_Bytes => Input_Bytes,
                  Output_Message => Output_Message,
                  Output_Bytes => Output_Bytes,
                  Max_Transmit => Max_Transmit,
                  Source_Address => Source_Address,
                  Domainname_Out => Domainname);
               IF Output_Message.Header.ANCOUNT > 0 or Output_message.Header.RCODE = DNS_Types.Not_Implemented THEN  --did we have the answer?
                  Output_Bytes := DNS_Types.message_length_range'Min(Max_Transmit,Output_Bytes);
                  --send non-authoritative response to requester here
                  --(non-authoritative bit already set by Create_Response)
                  --# accept Flow, 10, Output_message, "don't really care that the message is network ordered";
                  DNS_Network.Send_DNS_Message_Downstream(
                     Message      => Output_Message,
                     Number_Bytes => Output_Bytes,
                     To_Address   => Source_Address,
                     Failure      => Failure);
                  --# end accept;
                  if Failure then
                     Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "send failed",0);
                  end if;
               ELSE --answer not in cache, must forward to upstream server
                  --put downstream query in the query buffer
                  --# accept Flow, 10, Failure, "This can't fail since buffer is not full";
                    --outgoing port = 0 because we don't yet know it
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "New request for: ",0);
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  domainName,48);
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "From: ",0);
                  dns_network.Print_Server_Info(Source_address);
               if check_patterns.suspicious_query(domainname) then
                   Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Blocked!",0);
                   Protected_Spark_Io_05.Spark_IO_PO.New_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
               else
                  Global_Buffers.QueryBuffer.Add(Input_Message, Input_Bytes,
                     Source_Address, 0, WhereAdded, Failure);
                  --# end accept;
                  PrepareUpstreamQuery(WhereAdded, Input_Message);
                  --send new query upstream
                   --# accept Flow, 10, Input_Message, "don't really care that the message is network ordered";
                  DNS_Network.Send_DNS_Message_Upstream(
                     Message      => Input_Message,
                     Number_Bytes => Input_Bytes,
                     Upstream_Address => Upstream_Server,
                     Outgoing_Port => PortNumber,
                     Failure => Failure);
                  --# end accept;
                  --now we have the port number, set it explictly
                  Global_Buffers.QueryBuffer.SetPortNumber(WhereAdded, PortNumber);
                  IF Failure THEN
                     Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Send of query to upstream server failed",0);
                  END IF;
               end if;
               END IF; --answer to query in cache or not
            ELSE --query buffer full, purge any stale queries
               Global_Buffers.QueryBuffer.PurgeExpiredQueries;
            end if;  --query buffer full or not
         ELSE --got a response instead of a query
            IF NOT ResponseBufferFull THEN  --if room, put it in the response buffer
               --for responses, we already have the port number from receive_dns_message above
               Global_Buffers.ResponseBuffer.Add(Input_Message, Input_Bytes,
                  Source_Address, PortNumber, WhereAdded, Failure);
   	        IF Failure AND (WhereAdded = WhereAdded) THEN
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
               Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
               "Response buffer full -- server being flooded?",0);
            END IF;
            ELSE --got a response, but response buffer full
               null; --just drop it on the floor
            END IF;
         END IF;
      end loop;
   end Udp_query_Task_type;
end Udp_query_task_pkg;
