----------------------------------------------------------------
-- IRONSIDES - DNS SERVER
--
-- By: Martin C. Carlisle and Barry S. Fagin
--     Department of Computer Science
--     United States Air Force Academy
--
-- This is free software; you can redistribute it and/or
-- modify without restriction.  We do ask that you please keep
-- the original author information, and clearly indicate if the
-- software has been modified.
--
-- This software is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
----------------------------------------------------------------
--with ada.Text_IO;
PACKAGE BODY Udp_Response_Task_Pkg

IS

   -----------------------
   -- Udp_response_Task --
   -----------------------
   Startup_Suspension : Ada.Synchronous_Task_Control.Suspension_Object;
   Response_Task : Udp_Response_Task_Type; --THIS IS THE RESPONSE TASK

   procedure StartResponseTask
   is
   begin
      ada.synchronous_task_control.Set_True(Startup_Suspension);
   end StartResponseTask;

   --uses global udp_socket declared in dns_network.adb
   TASK BODY Udp_Response_Task_Type IS
      SlotOfResponse : Buffer_Pkg.BufferIndexType;
      SlotOfOriginalQuery: Buffer_Pkg.BufferIndexWithZeroType;
      Response : Dns_Types.dns_message;
      OriginalQuery: Dns_Types.dns_message;
      OriginalRequester : Dns_Network.Network_Address_And_Port;
      OriginalRandomizedName : RR_Type.WireStringType;
      OriginalName : RR_Type.WireStringType;
      ResponseName : RR_Type.WireStringType;
      OriginalPortNumber : Integer;
      ResponsePortNumber : Integer;
      NumBytes : DNS_Types.Message_Length_Range;
      Failure : Boolean;
      debug : boolean;
   BEGIN
      --block here until released by query task
      Ada.Synchronous_Task_Control.Suspend_Until_True(Startup_Suspension);

      loop
         --# assert true;

         --blocks if ResponseBuffer is empty
         Global_Buffers.ResponseBuffer.NextOccupiedSlot(SlotOfResponse);

         Response := Global_Buffers.ResponseBuffer.EntryAtSlot(
            SlotofResponse).Msg;
         NumBytes := Global_Buffers.ResponseBuffer.EntryAtSlot(
            SlotofResponse).NumBytes;

         SlotOfOriginalQuery :=
            Global_Buffers.QueryBuffer.FindSlotOfQueryWithConnectingID(Response.Header.MessageID);

         --returned value of 0 indicates no connecting query found, so the ID from upstream
         --query is invalid (possible cache poisoning attempt)
         IF SlotOfOriginalQuery = 0 THEN
            debug := Global_Buffers.QueryBuffer.Debugging;
            if debug then
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  "no query found for response with ID ",0);
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  Integer(Response.Header.MessageID),3,10);
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  ", ignored",0);
            end if;
            Global_Buffers.ResponseBuffer.Delete(SlotOfResponse, Failure);
            IF Failure THEN
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  "Attempt to delete from response buffer failed",0);
            END IF;
         ELSE --response linked to matching query
            OriginalQuery:= Global_Buffers.QueryBuffer.EntryAtSlot(
               SlotOfOriginalQuery).Msg;
            OriginalRequester := Global_Buffers.QueryBuffer.EntryAtSlot(
               SlotOfOriginalQuery).Requester;
            OriginalName := Global_Buffers.QueryBuffer.EntryAtSlot(
               SlotOfOriginalQuery).OriginalCaseDomainName;

            --get the query names from original query and response
            --(case is significant)
            OriginalRandomizedName := Global_Buffers.QueryBuffer.EntryAtSlot(
               SlotOfOriginalQuery).RandomizedCaseDomainName;
            ResponseName := Process_Dns_Request.Get_Query_Name(Response);

            --get the port numbers from original query and response
            OriginalPortNumber := Global_Buffers.QueryBuffer.EntryAtSlot(
               SlotofOriginalQuery).PortNumber;
            ResponsePortNumber := Global_Buffers.ResponseBuffer.EntryAtSlot(
               SlotOfResponse).PortNumber;

            --check to be sure casing of query names matches
            IF OriginalRandomizedName /= ResponseName THEN
               debug := Global_Buffers.ResponseBuffer.Debugging;
               if debug then
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                    Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                    OriginalRandomizedName ,0);
                  Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_String(
                    Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                    ResponseName,0);
                  Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                        "case mismatch in response to upstream query, ignored.",0);
               end if;
               Global_Buffers.ResponseBuffer.Delete(SlotOfResponse, Failure);
               IF Failure THEN
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Attempt to delete from response buffer failed",0);
               END IF;
               Global_Buffers.QueryBuffer.Delete(SlotOfOriginalQuery, Failure);
               IF Failure THEN
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Attempt to delete from query buffer failed",0);
               END IF;

            --check to be sure port numbers match
            ELSIF OriginalPortNumber /= ResponsePortNumber THEN
               debug := Global_Buffers.ResponseBuffer.Debugging;
               if debug then
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
                    Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                    OriginalPortNumber ,0,10);
                  Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Integer(
                    Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                    ResponsePortNumber,0,10);
                  Protected_SPARK_IO_05.SPARK_IO_PO.New_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,1);
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                        "port mismatch in response to upstream query, ignored.",0);
               end if;
               Global_Buffers.ResponseBuffer.Delete(SlotOfResponse, Failure);
               IF Failure THEN
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Attempt to delete from response buffer failed",0);
               END IF;
               Global_Buffers.QueryBuffer.Delete(SlotOfOriginalQuery, Failure);
               IF Failure THEN
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Attempt to delete from query buffer failed",0);
               END IF;

            ELSIF OriginalQuery.Header = DNS_Types.Empty_Header THEN --should never happen
               Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                  Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                  "no matching query for response returned from upstream server, ignored",0);

            --everything is in order, original query found that matches response
            ELSE
               --add the response to the DNS table and send answer back downstream
               Response.Header.MessageID := OriginalQuery.Header.MessageID;
               --test on NumBytes is to meet precondition for Send_dns_message
               IF Integer(NumBytes) > DNS_Types.Header_Bits/8 THEN
                  wire_to_records_pkg.Add_Records_To_Table(Response, NumBytes);
                  --above must be done before calling Send_dns_message,
                  --because Send_dns_message might scramble Response.header,
                  --depending on local byte ordering

                  --first restore original case of domain name
                  Process_Dns_Request.Set_Query_Name(
                     Input_Message => Response,
                     DomainName => OriginalName);
                  --and send the answer back downstream
                  DNS_Network.Send_DNS_Message_Downstream(
                     Message => Response,
                     Number_Bytes => NumBytes,
                     To_Address => OriginalRequester,
                     Failure => Failure);
               ELSE
                  Failure := true;
               END IF;
               IF Failure and --extra test makes flow error go away
                  Response.Header.QR = Response.Header.QR THEN
                  Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                     Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                     "Send of query to downstream server failed/insufficient bits received",0);
               ELSE
                  Global_Buffers.ResponseBuffer.Delete(SlotOfResponse, Failure);
                  IF Failure THEN
                     Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                        Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                        "Attempt to delete from response buffer failed",0);
                  END IF;
                  Global_Buffers.QueryBuffer.Delete(SlotOfOriginalQuery, Failure);
                  IF Failure THEN
                     Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(
                        Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,
                        "Attempt to delete from query buffer failed",0);
                  END IF;
               END IF; --successful transmission of response downstream
            END IF; --successful retreival of original query from buffer
         END IF; --upstream server returned valid response id
      end loop;
   end Udp_response_Task_type;
end Udp_response_task_pkg;
