SEPARATE(Wire_To_Records_Pkg)

   PROCEDURE Add_Records_To_Table (Response_Message: IN OUT Dns_Types.DNS_Message;
      Response_Bytes   : IN DNS_Types.message_length_range)
   is
      Domainname : RR_Type.WireStringType;
      --last byte of question section
      Query_End_Byte : DNS_Types.message_bytes_range;

      --walks through the records in the response
      BytePtr : DNS_Types.message_bytes_range;

      --how many records will be built from this packet
      NumberOfRecords : Natural;
      NameByte : Dns_Types.Byte;

      --field values for record currently being processed
      Record_Type : DNS_Types.Query_Type;
      Record_Class: DNS_Types.Query_Class;
      TTL : Unsigned_Types.Unsigned32;
      RDLength : Unsigned_Types.Unsigned16;
      ValidRecord : boolean;
      Ignored : boolean;

      PROCEDURE PreprocessResponse(Response_Message: IN Dns_Types.DNS_Message;
         NumberOfRecords : OUT Natural;
         Domainname : OUT RR_Type.WireStringType;
         BytePtr : OUT DNS_Types.Message_Bytes_Range;
         ValidRecord : OUT Boolean)
      --# derives NumberOfRecords from Response_Message &
      --#    Domainname from Response_Message &
      --#    BytePtr from Response_Message &
      --#    ValidRecord from Response_Message;
      --# post BytePtr + 12 <= dns_types.Message_Bytes_Range'last or
      --#    ValidRecord = false;
      IS
         Query_End_Byte : DNS_Types.message_bytes_range;
      BEGIN
         validRecord := true;
         --pull number of records (answer count) from HEADER section
         IF Response_Message.Header.ANCOUNT >
            dns_types.Unsigned_Short(Rr_Type.ReturnedRecordsIndexType'Last) THEN
               --more records than a bucket will hold?
               NumberOfRecords := Rr_Type.ReturnedRecordsIndexType'Last;
               --eventually need to add some logging/printing/warning msg here
         ELSE
            NumberOfRecords := Natural(Response_Message.Header.ANCOUNT);
         END IF;
         --Ada.Text_IO.Put("NumberOfRecords = " & Integer'Image(NumberOfRecords));
         --ada.Text_IO.New_Line;

          --extract info from the QUESTION section
         Get_Query_Name_End_Byte(Response_Message, Dns_Types.Message_Bytes_Range'First,
            Domainname, Query_End_Byte);
         --ensure you have enough room to move into the ANSWER section
         if Query_End_Byte >= dns_types.message_bytes_range'last then
            BytePtr := 1; --must be assigned, otherwise flow error
            validRecord := false;
         else
            BytePtr := Query_End_Byte+1;
         END IF;

         --ensure you have enough space for:
         -- NAME field (2 bytes if PTR, variable if inline FQDN)
         -- TYPE field 2 bytes
         -- CLASS field 2 bytes
         -- TTL field 4 bytes
         -- RDLENGTH field 2 bytes
         if BytePtr+12 > dns_types.message_bytes_range'last then
            validRecord := false;
         end if;
      END PreprocessResponse;

      PROCEDURE UpdateBytePtr(RDLength : IN Unsigned_Types.Unsigned16;
         BytePtr : IN OUT DNS_Types.Message_Bytes_Range;
         ValidRecord : IN OUT Boolean)
      --# derives BytePtr from RDLength, BytePtr &
      --#    ValidRecord from ValidRecord, RDLength, BytePtr;
      --# post BytePtr + 12 <= dns_types.Message_Bytes_Range'last or
      --#    ValidRecord = false;
      IS
      BEGIN
         --bad dynamic RDLENGTH?
         IF RDLength >= 1 and RDLength <= unsigned_types.Unsigned16(dns_types.message_bytes_range'last) THEN
            IF BytePtr + dns_types.message_bytes_range(RDLength) > dns_types.message_bytes_range'last THEN
               validRecord := false;
            ELSE
               BytePtr := BytePtr + Dns_Types.Message_Bytes_Range(RDLength);
               validRecord := validRecord; --avoid flow error
            END IF;
         ELSE
            ValidRecord := False;
         END IF;

         --bad packet?
         if BytePtr+12 > dns_types.message_bytes_range'last then
            validRecord := false;
         END IF;
      END UpdateBytePtr;

   BEGIN  --AddRecordsToTable
      PreprocessResponse(Response_Message => Response_Message,
         NumberOfRecords => NumberOfRecords,
         Domainname => Domainname,
         BytePtr => BytePtr,
         ValidRecord => ValidRecord);

      --build the records one at a time from the ANSWER section
      FOR Ctr IN Integer RANGE 1..NumberOfRecords LOOP
         exit when validRecord = false;
         --# assert BytePtr >= 1 and BytePtr + 12 <= dns_types.message_bytes_range'last;

         --first figure out if NAME field is 1 or 2 bytes
         NameByte:= Response_Message.Bytes(BytePtr);
         --Ada.Text_IO.Put("NameByte = " & Dns_Types.Byte'Image(NameByte));
         --ada.Text_IO.New_Line;

         --upper two bits = 00 --> FQDN is present
         IF NameByte <= DNS_Types.Byte(2**6-1) THEN
            --getting Domainname instead of query name, same code works for both
            Get_Query_Name_End_Byte(Response_Message, BytePtr, DomainName,
               Query_End_Byte);
            if Query_End_Byte >= dns_types.message_bytes_range'last-12 then
               BytePtr := 1; --must be assigned, otherwise flow error
               validRecord := false;
            else
               BytePtr := Query_End_Byte+1;
            END IF;

         --upper two bits = 01 --> EDNS0 format, NAME field is 1 byte
         ELSIF NameByte <= Dns_Types.Byte(2**7-1) THEN
            BytePtr := BytePtr + 1; --NEED TO DO SOMETHING ELSE HERE?

         --upper two bits = 11 --> PTR format, NAME field is 2 bytes
         ELSIF NameByte >= Dns_Types.Byte(2**7 + 2**6) THEN
            MakePtrFreeDomainNameFromWire(Response_Message, BytePtr, DomainName, validRecord);
            --rr_type.ConvertWireToDomainName(DomainName,StringDomainName,validRecord);
            --Ada.Text_IO.Put("Domain name = " & StringDomainName);
            --ada.Text_IO.New_Line;
            BytePtr := BytePtr + 2;
         --upper two bits = 10, invalid
         ELSE
            validRecord := false;
         END IF;

         --get the type and class of the record to be built
         Record_Type := Get_Record_Type(Message => Response_Message,
            BytePtr => BytePtr);
         BytePtr := BytePtr+2;
         --Ada.Text_IO.Put("   type = " & dns_types.query_Type'Image(Record_Type));
         --Ada.Text_IO.New_Line;
         Record_Class := Get_Record_Class(Message => Response_Message,
            BytePtr => BytePtr);
         BytePtr := BytePtr+2;
         --Ada.Text_IO.Put("   class = " & dns_types.query_class'Image(Record_class));
         --ada.Text_IO.New_Line;

         --TTL (4 bytes)
         Ttl := get_unsigned_32(Response_Message.Bytes,BytePtr);
         BytePtr := BytePtr+4;
         --Ada.Text_IO.Put("   TTL = " & unsigned_types.unsigned32'image(TTL));
         --Ada.Text_IO.New_Line;

         --RDLENGTH (2 bytes)
         RDLength := Unsigned_Types.Unsigned16(Response_Message.Bytes(BytePtr))*256 +
           Unsigned_Types.Unsigned16(Response_Message.Bytes(BytePtr+1));
         --RDLength has to be nonzero, but can't be larger than a DNS packet
         if RDLength = 0 or RDLength > unsigned_types.Unsigned16(dns_types.message_bytes_range'last) then
            ValidRecord := false;
         end if;
         BytePtr := BytePtr + 2;
         --Ada.Text_IO.Put("   RDLength = " & unsigned_types.Unsigned16'image(RDlength));
         --Ada.Text_IO.New_Line;

         --Byte is now at start of the RDATA field (record-dependent)
         --ready to process the RDATA section, build the record, and add it to the cache
         IF ValidRecord THEN
            Build_And_Insert_Record(Response_Message, Response_Bytes,
               Record_Type, BytePtr, Record_Class, TTL, RDLength,
               Domainname, ValidRecord);
         END IF;
         --check for bad RDLength here, update BytePtr if OK
         UpdateBytePtr(RDLength, BytePtr, ValidRecord);

         if not ValidRecord then
            Protected_SPARK_IO_05.SPARK_IO_PO.Put_Line(Protected_SPARK_IO_05.SPARK_IO_PO.Standard_Output,"invalid DNS msg received, ignored", 0);
      	 end if;  --maybe at some point add code to indicate where the bad msg came from
      END LOOP; --record processing

      --if response was NXDOMAIN, the RCODE of the header will be NAME_ERROR,
      --NumberOfRecords will be zero and record processing loop above will
      --not execute
      IF Response_Message.Header.RCODE = Dns_Types.Name_Error THEN
         --give the NXDOMAIN record a TTL of 15 minutes.  No data, so RDLENGTH=0.
         --# accept Flow, 10, Ignored,  "don't need this value";
         --# accept Flow, 23, Ignored,  "NXDOMAIN records are always valid";
         Build_And_Insert_Record(Response_Message, Response_Bytes, Dns_Types.NXDOMAIN,
            BytePtr, DNS_Types.NONE_CLASS, 900, 0, Domainname, Ignored);
      END IF;
      --what about ADDITIONAL and AUTHORITY sections?
   END Add_Records_To_Table;
