SEPARATE(Wire_To_Records_Pkg)
   PROCEDURE Build_RRSIG_Record(Packet : IN Dns_Types.DNS_Message;
      RDATAStartByte : IN Dns_Types.message_bytes_range;
      RDLength: IN Unsigned_Types.Unsigned16;  --need this to determine how long the sig is
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.RRSIG_Record_Type.RRSIGRecordType;
      Success : OUT Boolean)
      IS
         Now : Ada.Real_Time.Time;
         BytePtr : Dns_Types.message_bytes_range;
         NameStartPtr : Dns_Types.message_bytes_range;
         WireString : rr_type.WireStringType;
         BytesRead : Unsigned_Types.Unsigned16;
         Index : Rr_Type.Dnskey_Record_Type.DNSKeyStringTypeIndex;
         UncheckedSignatureLength : Unsigned_Types.Unsigned16;
         SignatureLength : Rr_Type.Rrsig_Record_Type.SigLengthValueType;
         SrcByte0: Dns_Types.Byte;
         SrcByte1: Dns_Types.Byte;
         srcByte2: dns_types.Byte;

      BEGIN
         ReturnedRec := rr_type.rrsig_record_type.blankRRSIGRecord; --pacify the SPARK monster
   	 ReturnedRec.TTLInSeconds := TTL;
         CASE Class IS
            when dns_types.IN_CLASS =>
               ReturnedRec.Class := Rr_Type.INTERNET;
            --sun will go nova before these get executed
      	    WHEN Dns_Types.CH_CLASS =>
               ReturnedRec.Class := Rr_Type.CH;
            WHEN Dns_Types.HS_CLASS =>
               ReturnedRec.Class := Rr_Type.HS;
            --SPARK ensures this never happens due to precondition
            WHEN Dns_Types.NONE_CLASS | Dns_Types.ANY_CLASS =>
               ReturnedRec.Class := Rr_Type.INTERNET;
         END CASE;
         BytePtr := RDATAStartByte;

         --TYPE COVERED (2 bytes)
         ReturnedRec.TypeCovered := Get_Record_Type(Packet, BytePtr);
         BytePtr := BytePtr + 2;

         --ALGORITHM (1 byte)
         ReturnedRec.Algorithm := unsigned_types.Unsigned8(Packet.Bytes(BytePtr));
         BytePtr := BytePtr + 1;

         --NUM LABELS (1 byte)
         ReturnedRec.NumLabels := unsigned_types.Unsigned8(Packet.Bytes(BytePtr));
         BytePtr := BytePtr + 1;

         --ORIG TTL (4 bytes)
         ReturnedRec.origTTL:= unsigned_types.Unsigned32(Packet.Bytes(BytePtr));
         BytePtr := BytePtr + 4;

         --SIG EXPIRATION (4 bytes)
         ReturnedRec.sigExpiration := unsigned_types.Unsigned32(Packet.Bytes(BytePtr));
         BytePtr := BytePtr + 4;

         --SIG INCEPTION (4 bytes)
         ReturnedRec.sigInception := unsigned_types.Unsigned32(Packet.Bytes(BytePtr));
         BytePtr := BytePtr + 4;

         --KEY TAG (2 bytes)
   	     ReturnedRec.keyTag := Get_Unsigned_16(Packet.Bytes, BytePtr);
         BytePtr := BytePtr + 2;

         --SIGNER NAME (var bytes)
         --copy the signer name from the wire to the record, expanding any pointers
         MakePtrFreeDomainNameFromWire(Packet, BytePtr, WireString, Success);
         IF Success THEN
            rr_type.ConvertWireToDomainName(WireString, ReturnedRec.SignerName, Success);
         END IF;

         --next we expect a valid signature field
         NameStartPtr := BytePtr;
         IF Success then
            FindFirstByteAfterName(Packet, NameStartPtr, BytePtr, Success);
         end if;
         --BytePtr now points to the first byte of the signature

         UncheckedSignatureLength := (RDLENGTH-18)-Unsigned_Types.Unsigned16(BytePtr-NameStartPtr);
         IF UncheckedSignatureLength < 3 OR UncheckedSignatureLength >
         Unsigned_Types.Unsigned16(Rr_Type.Rrsig_Record_Type.SigLengthValueType'Last) THEN
            Success := False;
         END IF;

         IF Success THEN
            --SIGNATURE (RDLENGTH-18-(BytePtr-NameStartPtr) bytes on the wire, longer when encoded)
            SignatureLength := integer(UncheckedSignatureLength);
            BytesRead := 0;
            Index := Rr_Type.RRSIG_Record_Type.RRSIGStringTypeIndex'First;

            --need to base64 encode what's on the wire
            WHILE BytesRead <= Unsigned_Types.Unsigned16(SignatureLength-3) AND
            BytePtr + 2 < Dns_Types.message_bytes_range'Last AND
            Index + 3 < Rr_Type.Dnskey_Record_Type.DNSKeyStringTypeIndex'Last loop
               --# assert BytesRead <= Unsigned_Types.Unsigned16(SignatureLength-3) AND
               --# BytePtr >= dns_types.message_bytes_range'First AND
               --# BytePtr + 2 < Dns_Types.message_bytes_range'Last AND
               --# Index >= Rr_Type.Dnskey_Record_Type.DNSKeyStringTypeIndex'First AND
               --# Index + 3 < Rr_Type.Dnskey_Record_Type.DNSKeyStringTypeIndex'Last AND
               --# SignatureLength >= 3 AND
               --# SignatureLength <= Rr_Type.Rrsig_Record_Type.SigLengthValueType'Last;

               --three bytes on wire
               SrcByte0 := Packet.Bytes(BytePtr);   --aaaaaabb
               SrcByte1 := Packet.Bytes(BytePtr+1); --bbbbcccc
               SrcByte2 := Packet.Bytes(BytePtr+2); --ccdddddd
               --four bytes in record
               ReturnedRec.signature(Index) := Base64ValToChar(SrcByte0/4); --aaaaaa
               ReturnedRec.signature(Index+1) := Base64ValToChar((SrcByte0 mod 4)*16 + SrcByte1/16); --bbbbbb
               ReturnedRec.signature(Index+2) := Base64ValToChar((SrcByte1 mod 16)*4 + SrcByte2/64); --cccccc
               ReturnedRec.signature(Index+3) := Base64ValToChar(SrcByte2 mod 64); --dddddd

               bytesRead:= bytesRead+3;
               BytePtr := BytePtr+3;
               Index := Index+4;
            END LOOP;

            --if number of bytes not divisible by 3, must handle as
            --described in http://en.wikipedia.org/wiki/Base64
            if Index+3 <= Rr_Type.Dnskey_Record_Type.DNSKeyStringTypeIndex'Last and
            BytePtr + 1 <= DNS_Types.message_bytes_range'Last THEN
               CASE (BytesRead-unsigned_types.Unsigned16(SignatureLength)) is
                  WHEN 0 => null;
                  WHEN 1 =>
                     SrcByte0 := Packet.Bytes(BytePtr);
                     SrcByte1 := 0;
                     ReturnedRec.signature(Index) := Base64ValToChar(SrcByte0/4); --aaaaaa
                     ReturnedRec.signature(Index+1) := Base64ValToChar((SrcByte0 mod 4)*16 + SrcByte1/16); --bbbbbb
                     ReturnedRec.signature(Index+2) := '=';
                     ReturnedRec.signature(Index+3) := '=';
                     BytesRead := BytesRead+1;
                  WHEN 2 => --some redundancy with above, but easier to read
                     SrcByte0 := Packet.Bytes(BytePtr);
                     SrcByte1 := Packet.Bytes(BytePtr+1);
                     SrcByte2 := 0;
                     ReturnedRec.signature(Index) := Base64ValToChar(SrcByte0/4); --aaaaaa
                     ReturnedRec.signature(Index+1) := Base64ValToChar((SrcByte0 mod 4)*16 + SrcByte1/16); --bbbbbb
                     ReturnedRec.signature(Index+2) := Base64ValToChar((SrcByte1 mod 16)*4 + SrcByte2/64); --cccccc
                     ReturnedRec.signature(Index+3) := '=';
                     BytesRead := BytesRead+2;
                  WHEN OTHERS => NULL;
               END CASE;
               ReturnedRec.SignatureLength := Index+3;
            END IF; --Index and BytePtr in bounds

            Now := Ada.Real_Time.Clock;
            ReturnedRec.TimeStamp := Now;

      	    --testing for validity here is slightly less efficient, but simplifies analysis

            --NOTE:  Not validating the algorithm for now.
      	    --Reference books and RFC say that 5 is the only valid value, but appears to be
   	        --an evolving standard.  berkeley.edu, for example, is already returning
   	        --records with an ALG field of 10 (SHA-2/512).  For now, just assume any value is OK, refine as
            --needed.  See http://www.iana.org/assignments/dns-sec-alg-numbers/dns-sec-alg-numbers.xhtml
   	        --for details.
            IF ReturnedRec.TypeCovered /= DNS_Types.UNIMPLEMENTED OR
            BytesRead /= unsigned_types.Unsigned16(SignatureLength) THEN
               Success := False;
            ELSE
               Success := True;
            END IF;
         END IF; --valid dynamic signature length
   end Build_RRSIG_Record;

