WITH Ada.Characters.Handling;
with flags;
PACKAGE BODY Wire_To_Records_Pkg IS
   function To_Lower(DomainName : in Rr_Type.WireStringType) return Rr_Type.WireStringType is
      LowerCaseVersion : Rr_Type.WireStringType := Rr_Type.BlankWire;
      length : rr_type.wireStringTypeIndex;
   begin
      Length := Rr_Type.WireNameLength(domainName);
      for I in Integer range 1..Length loop
         --# assert true;
         LowerCaseVersion(I) := Ada.Characters.Handling.to_lower(domainName(I));
      end loop;
      return lowerCaseVersion;
   END To_Lower;

   function Get_Unsigned_16(Bytes : IN Dns_Types.Bytes_Array_Type;
      StartByte : IN Dns_Types.message_bytes_range)
      RETURN Unsigned_Types.Unsigned16
      IS SEPARATE;

   function Get_Unsigned_32(Bytes : IN Dns_Types.Bytes_Array_Type;
      StartByte : IN Dns_Types.Message_Bytes_Range)
      return unsigned_types.Unsigned32
      IS SEPARATE;

   FUNCTION Get_Record_Type(Message: IN Dns_Types.DNS_Message;
      BytePtr : IN DNS_Types.Message_Bytes_Range)
      RETURN DNS_Types.Query_Type
   --# pre BytePtr + 12 <= dns_types.Message_Bytes_Range'last;
   IS
      RT_Natural : Natural;
      Record_Type : Dns_Types.Query_Type;
      FUNCTION To_Type IS NEW Ada.Unchecked_Conversion(
         natural,dns_types.query_type);
      FUNCTION Type_To_Natural IS NEW Ada.Unchecked_Conversion(
         Dns_Types.Query_Type,Natural);
   BEGIN
      --TYPE field (2 bytes, encoded same as query type)
      RT_Natural := Natural(Message.Bytes(BytePtr))*256 +
         Natural(Message.bytes(BytePtr+1));
      --# accept Warning, 12, type_To_Natural, "unchecked conversions ok";
      IF RT_Natural >= Type_To_Natural(DNS_Types.Query_Type'First) AND
         RT_Natural <= type_To_Natural(DNS_Types.Query_Type'Last) then
         --# end accept;
         --# accept Warning, 12, To_Type, "unchecked conversions ok";
         Record_Type := To_Type(RT_Natural);
         --# end accept;
         if not Record_Type'Valid then
            Record_Type := DNS_Types.UNIMPLEMENTED;
         end if;
      else
         Record_Type := DNS_Types.UNIMPLEMENTED;
      END IF;
      Return Record_Type;
   END Get_Record_Type;

   FUNCTION Get_Record_Class(Message: IN Dns_Types.DNS_Message;
      BytePtr : IN DNS_Types.Message_Bytes_Range)
      RETURN DNS_Types.Query_Class
   --# pre BytePtr + 12 <= dns_types.Message_Bytes_Range'last;
   IS
      RC_Natural : Natural;
      Record_Class : Dns_Types.Query_Class;
      FUNCTION To_Class IS NEW Ada.Unchecked_Conversion(
         natural,dns_types.query_class);
      FUNCTION Class_to_Natural IS NEW Ada.Unchecked_Conversion(
         Dns_Types.query_class,Natural);
   BEGIN
      --CLASS field (2 bytes)
      RC_Natural := Natural(Message.Bytes(BytePtr))*256 +
         Natural(Message.bytes(BytePtr+1));
      --# accept Warning, 12, class_To_Natural, "unchecked conversions ok";
      IF RC_Natural >= Class_To_Natural(DNS_Types.Query_Class'First) AND
         RC_Natural <= class_to_natural(DNS_Types.Query_Class'Last) then
         --# end accept;
         --# accept Warning, 12, To_Class, "unchecked conversions ok";
         Record_Class := To_Class(RC_Natural);
         --# end accept;
         if not Record_Class'Valid then
            Record_Class := DNS_Types.NONE_CLASS;
         end if;
      else
         Record_Class := DNS_Types.NONE_CLASS;
      END IF;
      Return Record_Class;
   END Get_Record_Class;


   FUNCTION IsLengthByte(Byte : IN Dns_Types.Byte) RETURN Boolean IS
      RetVal : boolean;
   BEGIN
      IF Byte <= Dns_Types.Byte(2**7-1) THEN
         RetVal := True;
      ELSE
         RetVal := False;
      END IF;
      return RetVal;
   END IsLengthByte;

   FUNCTION IsPtrByte(Byte : IN Dns_Types.Byte) RETURN Boolean IS
   RetVal : Boolean;
   BEGIN
      IF Byte >= dns_types.byte(2**7 + 2**6) THEN
         RetVal := True;
      ELSE
         RetVal := False;
      END IF;
      return RetVal;
   END IsPtrByte;

   --implements Base64 index table at http://en.wikipedia.org/wiki/Base64
   --Arg is the value, ans is the character
   FUNCTION Base64ValToChar(Arg: IN DNS_Types.Byte) RETURN Character IS
      ans : Character;
   BEGIN
      CASE Arg IS
         WHEN 0..25 => ans := Character'Val(Arg + Character'Pos('A'));
         WHEN 26..51 => ans := Character'Val((Arg - 26) + Character'Pos('a'));
         WHEN 52..61 => ans := Character'Val((Arg - 52) + Character'Pos('0'));
         WHEN 62 => ans := '+';
         WHEN 63 => ans := '/';
         WHEN OTHERS => ans := ' ';
      end case;
      return ans;
   END Base64ValToChar;

   PROCEDURE FindFirstByteAfterName(Packet : IN Dns_Types.DNS_Message;
      NameStartPtr : IN Dns_Types.Message_Bytes_Range;
      BytePtr: out Dns_Types.Message_Bytes_Range;
      Success : OUT Boolean)
      --# derives BytePtr from Packet, NameStartPtr &
      --#    Success from Packet, NameStartPtr;
      --# post BytePtr >= NameStartPtr;
   IS
   CurrentByte : DNS_types.Byte;
   BEGIN
      Success := True;
      --the rest of the data starts after the next PTR field (2 bytes) or the next
      --zero byte (1 byte) after NameStartByte
      BytePtr := NameStartPtr;
      CurrentByte := Packet.Bytes(BytePtr); --avoid flow error
      WHILE BytePtr < Dns_Types.Message_Bytes_Range'Last LOOP
         --#assert BytePtr >= NameStartPtr and
         --#   BytePtr < Dns_Types.Message_Bytes_Range'Last;
         CurrentByte := Packet.Bytes(BytePtr);
         exit when CurrentByte = 0 or IsPtrByte(CurrentByte);
         BytePtr := BytePtr+1;
      END LOOP;

      IF BytePtr = Dns_Types.Message_Bytes_Range'Last THEN
         Success := False;
      ELSIF CurrentByte = 0 then
         BytePtr := BytePtr + 1;
      ELSE --must be a PTR field
         IF BytePtr >= Dns_Types.Message_Bytes_Range'Last-1 THEN
            Success := False;
         ELSE
            BytePtr := BytePtr + 2;
         END IF;
      END IF;
   END FindFirstByteAfterName;

   procedure MakePtrFreeDomainNameFromWire(Packet : IN Dns_Types.DNS_Message;
      NameStartByte : IN Dns_Types.Message_Bytes_Range;
      PtrFreeDomainName : OUT Rr_Type.WireStringType;
      Success : OUT boolean )
   --# derives PtrFreeDomainName from Packet, NameStartByte &
   --#    Success from Packet, NameStartByte;
   IS
      LengthOrPtrByte : Dns_Types.Byte;
      Index : Rr_Type.WireStringTypeIndex;
      BytesCopied : Unsigned_Types.Unsigned16;
      BytesToCopy: Unsigned_Types.Unsigned16;
      BytePtr: Dns_Types.Message_Bytes_Range;

      --we need to limit the number of ptr fields processed in a name to ensure that badly
      --circular pointer chains in a bad/malicious packet don't cause an infinite loop
      MAX_NUM_PTR_FIELDS : CONSTANT Integer := 8;

      PtrFieldCtr : Integer;
      Offset : Unsigned_Types.Unsigned16; --temp value, used to check for valid offsets

   BEGIN
      Success := True;
      PtrFreeDomainName := Rr_Type.BlankWire;
      Index := rr_type.WireStringTypeIndex'First;
      BytesCopied := 0;
      BytesToCopy := 0;
      PtrFieldCtr := 0;
      BytePtr := NameStartByte;
      LengthOrPtrByte := Packet.Bytes(BytePtr);

      WHILE LengthOrPtrByte /= 0 AND
      BytesCopied < Unsigned_Types.Unsigned16(Rr_Type.WireStringTypeIndex'Last) AND
      PtrFieldCtr <= MAX_NUM_PTR_FIELDS AND Success LOOP
         --# assert BytesCopied >= 0 and BytesCopied < unsigned_types.Unsigned16(rr_type.WireStringTypeIndex'last) and
         --# BytePtr >= 0 and BytePtr <= Dns_Types.Message_Bytes_Range'Last and
         --# PtrFieldCtr <= MAX_NUM_PTR_FIELDS and
         --# unsigned_types.unsigned16(index) = BytesCopied + 1;
         IF BytesToCopy = 0 THEN
            LengthOrPtrByte := Packet.Bytes(BytePtr);
            IF IsLengthByte(LengthOrPtrByte) THEN
               BytesToCopy := Unsigned_types.Unsigned16(LengthOrPtrByte)+1; --must also copy the length byte itself
            ELSIF IsPtrByte(LengthOrPtrByte) THEN
               PtrFieldCtr := PtrFieldCtr + 1;
               --offset is lower 14 bits of two-byte field
               Offset := Get_Unsigned_16(Packet.Bytes,BytePtr) - Unsigned_Types.Unsigned16(16#C000#);
               --name pointers are offsets into the entire message, so must
               --subtract the header size in bytes, +1 because array is indexed from 1
               Offset := (Offset-dns_types.header_bytes)+1;
               if Offset = 0 or Offset >  unsigned_types.Unsigned16(Dns_Types.Message_Bytes_Range'Last) THEN
                  Success := False;
               ELSE
                  BytePtr := Dns_Types.Message_Bytes_Range(Offset);
                  BytesToCopy := 0;
               end if;
            ELSE
               Success := False;
            END IF;
         END IF;
         IF BytesToCopy > 0 then
            PtrFreeDomainName(Index) := Character'val(Packet.Bytes(BytePtr));
            BytesCopied := BytesCopied + 1;
            Offset := Unsigned_Types.Unsigned16(BytePtr);
            IF Offset = Unsigned_Types.Unsigned16(Dns_Types.Message_Bytes_Range'Last) THEN
               Success := False;
            else
               BytePtr := BytePtr + 1;
            end if;
            IF BytesCopied < unsigned_types.Unsigned16(rr_type.WireStringTypeIndex'last) THEN
               Index := Index + 1;
            end if;
            BytesToCopy := BytesToCopy - 1;
         end if;
      END LOOP;

      IF PtrFieldCtr > MAX_NUM_PTR_FIELDS THEN
         Success := False;
      END IF;

   END MakePtrFreeDomainNameFromWire;

   procedure Get_Mail_Exchanger (
      Input_Packet  : in DNS_Types.DNS_Message;
      Input_Bytes   : in DNS_Types.message_length_range;
      Start_Byte    : in DNS_types.Message_Bytes_Range;
      Domainname    : out RR_Type.wireStringType)
      --# derives Domainname
      --#            from Input_Bytes,
      --#                         Input_Packet,
      --#			  Start_Byte;
   	  --  packet must at least contain enough bytes for a header, min size domain name,
   	  --  type and class (1 + 2 + 2 bytes)
      --# pre Input_Bytes >=DNS_Types.Header_Bits/8+5 and Start_Byte >= 1;
   is separate;

   --methods for building specific records from a DNS packet off the wire

   --If it can be determined before the method is called whether or not
   --the record on the wire is valid, use a function that returns a record
   --of the appropriate type.  If not, use a procedure that has the record
   --and a boolean as out parameters.

   FUNCTION Build_A_Record(Packet : IN Dns_Types.DNS_Message;
                           RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
                           TTL : IN Unsigned_Types.Unsigned32;
                           Class : IN dns_types.Query_Class)
                           RETURN Rr_Type.A_Record_Type.ARecordType
   --# global in Ada.Real_Time.ClockTime;
   --RDATA section must contain at least 4 bytes to be valid
   --# pre RDATAStartByte <= Dns_Types.Message_Bytes_Range'last-3;
                           is separate;


   FUNCTION Build_Corrupt_Record(Packet : IN Dns_Types.DNS_Message;
                           RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
                           TTL : IN Unsigned_Types.Unsigned32;
                           Class : IN dns_types.Query_Class)
                              RETURN Rr_Type.Corrupt_Record_Type.CorruptRecordType
   --# global in Ada.Real_Time.ClockTime;
   --RDATA section must contain at least 4 bytes to be valid
   --# pre RDATAStartByte <= Dns_Types.Message_Bytes_Range'last-3;
                              is separate;


   FUNCTION Build_AAAA_Record(Packet : IN Dns_Types.DNS_Message;
                              RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
                              TTL : IN Unsigned_Types.Unsigned32;
                              Class : IN dns_types.Query_Class)
                              RETURN Rr_Type.AAAA_Record_Type.AAAARecordType
   --# global in Ada.Real_Time.ClockTime;
   --RDATA section must contain at least 16 bytes to be valid
   --# pre RDATAStartByte <= Dns_Types.Message_Bytes_Range'last-15;
      IS SEPARATE;

   PROCEDURE Build_CNAME_Record(Packet : IN Dns_Types.DNS_Message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.CNAME_Record_Type.CNAMERecordType;
      Success : OUT boolean)
   --# global in Ada.Real_Time.ClockTime;
   --# derives ReturnedRec from Packet, RDATAStartByte, TTL, CLass, Ada.Real_Time.ClockTime &
   --#    Success from Packet, RDATAStartByte;
   IS SEPARATE;

   FUNCTION Build_MX_Record(Packet : IN Dns_Types.DNS_Message;
      ResponseBytes   : IN DNS_Types.Message_Length_Range;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN dns_types.Query_Class)
      RETURN Rr_Type.MX_Record_Type.MXRecordType
   --# global in Ada.Real_Time.ClockTime;
   --RDATA section must contain at least 6 bytes to be valid
   --# pre RDATAStartByte <= Dns_Types.Message_Bytes_Range'last-5 and ResponseBytes >=DNS_Types.Header_Bits/8+5;
   IS SEPARATE;

   PROCEDURE Build_PTR_Record(Packet : IN Dns_Types.DNS_Message;
         RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
         TTL : IN Unsigned_Types.Unsigned32;
         Class : IN Dns_Types.Query_Class;
         ReturnedRec : OUT Rr_Type.PTR_Record_Type.PTRRecordType;
         Success : OUT boolean)
   --# global in Ada.Real_Time.ClockTime;
   --# derives ReturnedRec from Packet, RDATAStartByte, TTL, CLass, Ada.Real_Time.ClockTime &
   --#    Success from Packet, RDATAStartByte;
   IS SEPARATE;

   PROCEDURE Build_NS_Record(Packet : IN Dns_Types.dns_message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.NS_Record_Type.NSRecordType;
      Success : OUT boolean)
   --# global in Ada.Real_Time.ClockTime;
   --# derives ReturnedRec from Packet, RDATAStartByte, TTL, CLass, Ada.Real_Time.ClockTime &
   --#    Success from Packet, RDATAStartByte;
   IS SEPARATE;

   FUNCTION Build_NXDOMAIN_Record(
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN dns_types.Query_Class)
      RETURN Rr_Type.NXDOMAIN_Record_Type.NXDOMAINRecordType
   --# global in Ada.Real_Time.ClockTime;
   IS SEPARATE;

   PROCEDURE Build_SOA_Record(Packet : IN Dns_Types.dns_message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.SOA_Record_Type.SOARecordType;
      Success : OUT boolean)
   --# global in Ada.Real_Time.ClockTime;
   --# derives ReturnedRec from Packet, RDATAStartByte, TTL, CLass, Ada.Real_Time.ClockTime &
   --#    Success from Packet, RDATAStartByte;
   IS SEPARATE;

--DNSSEC
   PROCEDURE Build_DNSKEY_Record(Packet : IN Dns_Types.dns_message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      RDLength : IN Unsigned_Types.Unsigned16;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN dns_types.Query_Class;
      ReturnedRec : OUT Rr_Type.dnskey_Record_Type.DNSKEYRecordType;
      Success : OUT boolean)
   --# global in Ada.Real_Time.ClockTime;
   --# derives ReturnedRec from Packet, RDATAStartByte, RDLength, TTL, CLass, Ada.Real_Time.ClockTime &
   --#    Success from Packet, RDATAStartByte, RDLength;
   --RDATA section must contain at least 4 bytes to be valid
   --# pre RDATAStartByte + 4 <= Dns_Types.Message_Bytes_Range'last;
   IS SEPARATE;

PROCEDURE Build_RRSIG_Record(Packet : IN Dns_Types.dns_message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      RDLength: IN Unsigned_Types.Unsigned16;  --need this to determine how long the sig is
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.RRSIG_Record_Type.RRSIGRecordType;
      Success : OUT Boolean)
      --# global in Ada.Real_Time.ClockTime;
      --# derives ReturnedRec from Packet, RDATAStartByte, RDLength, TTL, CLass, Ada.Real_Time.ClockTime &
      --#    Success from Packet, RDATAStartByte, RDLength;
      --RDATA section must contain at least 18 bytes to be valid
      --# pre RDATAStartByte +18 <= Dns_Types.Message_Bytes_Range'last;
   IS SEPARATE;

PROCEDURE Build_NSEC_Record(Packet : IN Dns_Types.dns_message;
      RDATAStartByte : IN Dns_Types.Message_Bytes_Range;
      RDLength: IN Unsigned_Types.Unsigned16;
      TTL : IN Unsigned_Types.Unsigned32;
      Class : IN Dns_Types.Query_Class;
      ReturnedRec : OUT Rr_Type.NSEC_Record_Type.NSECRecordType;
      Success : OUT Boolean)
      --# global in Ada.Real_Time.ClockTime;
      --# derives ReturnedRec from Packet, RDATAStartByte, RDLength, TTL, CLass, Ada.Real_Time.ClockTime &
      --#    Success from Packet, RDATAStartByte, RDLength;
      --# pre RDLength >= 2 and RDLength <= unsigned_types.unsigned16(dns_types.Message_Bytes_Range'last);
      --# post Success = False or ReturnedRec.NumberOfBlocks > 0;
   IS SEPARATE;

--a dispatcher to the methods above
PROCEDURE Build_And_Insert_Record(
      Response_Message: IN OUT Dns_Types.DNS_Message; --the response we got from the wire
      Response_Bytes: IN DNS_Types.message_length_range;
      Record_Type : IN DNS_Types.Query_Type;  --the type of record to be built/added
      BytePtr : IN Dns_Types.message_bytes_range; --first byte of RDATA section
      Record_Class: IN DNS_Types.Query_Class;
      TTL : IN Unsigned_Types.Unsigned32;
      RDLength : IN Unsigned_Types.Unsigned16;
      Domainname : IN RR_Type.WireStringType;
      ValidRecord : IN OUT Boolean)
      --# global in Ada.Real_Time.ClockTime;
      --#    in out dns_table_pkg.dns_table, protected_spark_io_05.spark_io_po;
      --# derives dns_table_pkg.dns_table from dns_table_pkg.dns_table,
      --#    Ada.Real_Time.ClockTime, Response_Message, Response_Bytes, Record_Type,
      --#    BytePtr, Record_Class, TTL, RDLength, Domainname &
      --#    protected_spark_io_05.spark_io_po from *, record_type, byteptr,
      --#    domainname, response_message &
      --#    ValidRecord from Response_Message, Response_Bytes, Record_Type, BytePtr, RDLength,
      --#    ValidRecord &
      --#    Response_Message from dns_table_pkg.dns_table;
      --# pre ((RDLength = 0 and Record_Type = dns_types.NXDOMAIN) or RDlength > 0)
      --#    and RDLength <= Unsigned_types.Unsigned16(dns_types.message_bytes_range'last);

      IS SEPARATE;

procedure Get_Query_Name_End_Byte (
      Input_Packet  : in DNS_Types.dns_message;
      Start_Byte    : IN Dns_Types.Message_Bytes_Range;
      Domainname    : OUT RR_Type.WireStringType;
      End_Byte      : out Dns_types.Message_Bytes_Range)
      --# derives Domainname, End_Byte from Input_Packet, Start_Byte;
      --# pre Start_Byte < dns_types.Message_Bytes_Range'last-4;
      --# post End_Byte >= Start_Byte+dns_types.Message_Bytes_Range(4) and
      --#    End_Byte <= Start_Byte+dns_types.Message_Bytes_Range(RR_Type.WireStringType'Last+4);
   is separate;

--had to make this separate because it became too big for the Examiner
PROCEDURE Add_Records_To_Table (Response_Message: IN OUT Dns_Types.DNS_Message;
   Response_Bytes   : IN DNS_Types.Message_Length_Range)
   is separate;
end wire_to_records_pkg;
